/*
 *	sd_grailpxm.cxx
 *
 *	CAUTION: For 32-bit Windows only !
 *
 *	GRAphical Interface Layer pixel map interface class. This is
 *	heavily used for double buffering.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "sd_grafix.h"

#include <OPaC/grail.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION (grail.h);
OP_IMPLEMENT_CLASS (OPGrailPixmap);

/****************************************************************/

/*
 *	Initialise the Grail Pixmap instance. This won't allocate
 *	any bitmap; do this using "Initialise" !
 */
	
OPGrailPixmap::OPGrailPixmap ()
{
	OP_METHOD (OPGrailPixmap, 0);
	
	this->dx     = 0;
	this->dy     = 0;
	this->depth  = 0;
}



/*
 *	Destroy cleanly a Pixmap : this will free an existing
 *	bitmap and associated data.
 */

OPGrailPixmap::~OPGrailPixmap ()
{
	OP_METHOD (~OPGrailPixmap, 0);
}



/*
 *	Start drawing into the window. This will first draw into
 *	the buffering pixmap.
 */

void
OPGrailPixmap::BeginDraw (OPRectIn box, OPRelCoord margin)
{
	OP_METHOD (BeginDraw, 2);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPRelCoord, margin);
	super::BeginDraw (box, margin);
}



/*
 *	Start drawing into the pixel map. Tell the Grail Port
 *	which display context to use for drawing.
 */

void
OPGrailPixmap::BeginDraw (OPCoord x1, OPCoord y1, OPCoord x2, OPCoord y2)
{
	OP_METHOD (BeginDraw, 4);
	OP_ARG (1, OPCoord, x1);
	OP_ARG (2, OPCoord, y1);
	OP_ARG (3, OPCoord, x2);
	OP_ARG (4, OPCoord, y2);
	super::BeginDraw (x1, y1, x2, y2);
}



/*
 *	End of drawing. Just propagate the message and remove
 *	the display context from the Grail Port.
 */

void
OPGrailPixmap::EndDraw ()
{
	OP_METHOD (EndDraw, 0);
	super::EndDraw ();
}



/*
 *	Initialise a new pixel map. If one already existed, just
 *	delete it and start anew. If "window" is not specified,
 *	the default screen will be used as model.
 */

void
OPGrailPixmap::Initialise (OPGrailPixmapIn window, OPCoord dx, OPCoord dy)
{
	OP_METHOD (Initialise, 3);
	OP_ARG (1, OPGrailPixmapIn, window);
	OP_ARG (2, OPCoord, dx);
	OP_ARG (3, OPCoord, dy);
	OP_ASSERT ((dx > 0) && (dy > 0));
	OP_ASSERT (this->is_drawing == FALSE);
	
	this->surface = GfxSurface::New (32);
	this->surface->Initialise ((int)dx, (int)dy);
	
	this->is_ready  = TRUE;
	this->dx        = dx;
	this->dy        = dy;
}



/*
 *	Initialise a new pixel map with the specified depth.
 *	If one already existed, just delete it and start anew.
 */

void
OPGrailPixmap::Initialise (Count depth, OPCoord dx, OPCoord dy)
{
	OP_METHOD (Initialise, 3);
	OP_ARG (1, Count, depth);
	OP_ARG (2, OPCoord, dx);
	OP_ARG (3, OPCoord, dy);
	
	OP_ASSERT ((dx > 0) && (dy > 0));
	OP_ASSERT (this->is_drawing == FALSE);
	
	this->surface = GfxSurface::New (32);
	this->surface->Initialise ((int)dx, (int)dy);
	
	this->is_ready  = TRUE;
	this->dx        = dx;
	this->dy        = dy;
	this->depth     = depth;
}


/*
 *	Return the RGB compontents of a few pixels by either reading back
 *	the cache (if any) or by using the system dependent implementation.
 */

Bool
OPGrailPixmap::GetRGBPixels (OPCoord x, OPCoord y, OPCoord dx, Card8* r, Card8* g, Card8* b)
{
	OP_METHOD (GetRGBPixels, 6);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, Card8*, r);
	OP_ARG (5, Card8*, g);
	OP_ARG (6, Card8*, b);

	Card32 num  = (Card32)(dx);
	Card32 line = (Card32)(y);
	Card32 pos  = (Card32)(x);
	
	while (num--) {
		Card32 rgb = this->surface->GetPixelRGB (pos++, line);
		*r++ = (Card8)(rgb >> 16);
		*g++ = (Card8)(rgb >>  8);
		*b++ = (Card8)(rgb >>  0);
	}
	
	return TRUE;
}

Bool
OPGrailPixmap::GetColorPixels (OPCoord x, OPCoord y, OPCoord dx, OPColorPixel* pixels)
{
	OP_METHOD (GetColorPixels, 4);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, OPColorPixel*, pixels);

	OP_ASSERT (this->surface);
	
	Card32 num  = (Card32)(dx);
	Card32 line = (Card32)(y);
	Card32 pos  = (Card32)(x);
	
	while (num--) {
		*pixels++ = this->surface->GetPixelRGB (pos++, line);
	}
	
	return TRUE;
}

OPColorPixel
OPGrailPixmap::ReturnPixel (OPCoord x, OPCoord y)
{
	OP_METHOD (ReturnPixel, 2);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	
	OP_ASSERT (this->surface);
	
	return this->surface->GetPixelRGB ((int)x, (int)y);
}

