/*
 *	sd_grailfnt.cxx
 *
 *	GRAphical Interface Layer Font representation. This is machine
 *	dependent. Implementation for X11.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

// Modified by Laurent Bovet 11.12.97

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "sd_interface.h"

#include <OPaC/database.h>
#include <OPaC/grail.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPGrailFont);

/****************************************************************/

OPGrailFont::OPGrailFont ()
{
	OP_METHOD (OPGrailFont, 0);
	
	this->info  = 0;
	this->name  = OPStringNull;
	this->style = OPStringNull;
	this->size  = 0;
}



OPGrailFont::~OPGrailFont ()
{
	OP_METHOD (~OPGrailFont, 0);
	
	
	if (this->info) {
		OP_FREE_FROM_ZONE (this->ReturnMemZone (), this->info);
	}
}



void
OPGrailFont::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	super::StoreInstance (storage);
	
	storage << this->name;
	storage << this->style;
	storage << this->size;
}



void
OPGrailFont::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	super::RestoreInstance (storage);
	
	storage >> this->name;
	storage >> this->style;
	storage >> this->size;
	
	this->RealizeFont ();
}



Bool
OPGrailFont::Initialise (OPStringIn name, OPStringIn style, OPCoord size)
{
	OP_METHOD (Initialise, 3);
	OP_ARG (1, OPStringIn, name);
	OP_ARG (2, OPStringIn, style);
	OP_ARG (3, OPCoord, size);
	
	if (this->info) {
		return FALSE;
	}
	
	if (this->name.IsValid () == FALSE) {
		this->name = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPString);
	}
	
	if (this->style.IsValid () == FALSE) {
		this->style = OP_NEW_FROM_ZONE (this->ReturnMemZone (), OPString);
	}
	
	this->name->SetString (name);
	this->style->SetString (style);
	
	this->size = size;
	
	return this->RealizeFont ();
}



Bool
OPGrailFont::RealizeFont ()
{
	OP_METHOD (RealizeFont, 0);
	
	if ( (this->name.IsValid ())
	  && (this->style.IsValid ())
	  && (this->size) ) {
		
		if (this->info == 0) {
			this->info = OP_ALLOC_FROM_ZONE (this->ReturnMemZone (), OPaC_GrFont);
			this->info->xfont = 0;
		}
		
		this->info->xfont = XLoadFont(xdpy, "Fixed");  // Default font
		
		return TRUE;
	}
	
	return FALSE;
}



void
OPGrailFont::GetFontName (OPStringOut name) const
{
	OP_METHOD (GetFontName, 1);
	OP_ARG (1, OPStringOut, name);
	
	name = this->name;
}



void
OPGrailFont::GetFontStyle (OPStringOut name) const
{
	OP_METHOD (GetFontStyle, 1);
	OP_ARG (1, OPStringOut, name);
	
	name = this->style;
}



OPCoord
OPGrailFont::ReturnFontSize () const
{
	OP_METHOD (ReturnFontSize, 0);
	return (unsigned int)( this->ReturnAscender() - this->ReturnDescender());
}



OPRelCoord
OPGrailFont::ReturnAscender () const
{
	OP_METHOD (ReturnAscender, 0);
	
	LatinChar *buffer = "dummy";

	XCharStruct properties; // overall characteristics of the string
	XGCValues values;
	int direction, ascent, descent;

	XQueryTextExtents( xdpy, this->info->xfont, buffer, strlen(buffer),
										 &direction, &ascent, &descent,
										 &properties); // Get the characteristics of the string


	return (OPRelCoord)ascent;
}



OPRelCoord
OPGrailFont::ReturnDescender () const
{
	OP_METHOD (ReturnDescender, 0);
	
	LatinChar *buffer = "dummy";

	XCharStruct properties; // overall characteristics of the string
	XGCValues values;
	int direction, ascent, descent;

	XQueryTextExtents( xdpy, this->info->xfont, buffer, strlen(buffer),
										 &direction, &ascent, &descent,
										 &properties); // Get the characteristics of the string


	return (OPRelCoord)- descent -1 ;
}



OPCoord
OPGrailFont::ReturnWidth (const LatinChar* text) const
{
	OP_METHOD (ReturnWidth, 1);
	OP_ARG (1, const LatinChar*, text);
	
	if ( (text == 0)
	  || (text[0] == 0) ) {
		return 0;
	}
	
	OPCoord width = 0;
	
	if ( (this->info != 0)
	  && (this->info->xfont != 0) ) {
		
		XCharStruct properties; // overall characteristics of the string
		XGCValues values;
		int direction, ascent, descent;
		
		XQueryTextExtents( xdpy, this->info->xfont, text, strlen(text),
											 &direction, &ascent, &descent,
											 &properties); // Get the characteristics of the string
		
		

		width = (OPCoord) properties.width;
	}
	
	return width;
}



OPCoord
OPGrailFont::ReturnWidth (OPStringIn text) const
{
	OP_METHOD (ReturnWidth, 1);
	OP_ARG (1, OPStringIn, text);
	
	if (text.IsValid () == FALSE) {
		return 0;
	}
	
	Count len = text->CountCharacters ();
	
	return this->ReturnWidth (text, 0, len);
}



OPCoord
OPGrailFont::ReturnWidth (OPStringIn text, Count start, Count length) const
{
	OP_METHOD (ReturnWidth, 3);
	OP_ARG (1, OPStringIn, text);
	OP_ARG (2, Count, start);
	OP_ARG (3, Count, length);
	
	if (text.IsValid () == FALSE) {
		return FALSE;
	}
	
	OPCoord width = 0;
	Count   len   = text->CountCharacters ();
	Count   pos   = start + length;
	
	if ( (this->info != 0)
	  && (this->info->xfont != 0)
	  && (length != 0)
	  && (start < len) ) {
		
		if (pos > len) length -= (pos - len);
		
		LatinChar buffer[text->CountCharacters()+1];
		
		text->ExtractStringCopy (buffer, start, length);
		
		width = this->ReturnWidth(buffer);
	}
	
	return width;
}
