/*
 *	sd_file_io.cxx
 *	
 *	MS-Windows 32-bit file input/output functions.
 *	
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "sd_interface.h"
#include <io.h>

#include <OPaC/string.h>
#include <OPaC/sd/file_io.h>

OPaC_FileDesc
OPaC_OpenFileRead (OPStringIn name)
{
	OP_FUNCTION (OPaC_OpenFileRead, 1);
	OP_ARG (1, OPStringIn, name);
	
	HFILE    hfile;
	OFSTRUCT info;
	UINT     mode = OF_READ;
	
	char buffer[250];
	name->GetStringCopy (buffer, 250);
	
	hfile = OpenFile (buffer, &info, mode);
	
	if (hfile == HFILE_ERROR)
		return 0;
	
	OP_ASSERT (hfile);
	
	return (OPaC_FileDesc)(hfile);
}


OPaC_FileDesc
OPaC_OpenFileWrite (OPStringIn name)
{
	OP_FUNCTION (OPaC_OpenFileWrite, 1);
	OP_ARG (1, OPStringIn, name);
	
	HFILE    hfile;
	OFSTRUCT info;
	UINT     mode = OF_WRITE | OF_CREATE;
	
	char buffer[250];
	name->GetStringCopy (buffer, 250);
	
	hfile = OpenFile (buffer, &info, mode);
	
	if (hfile == HFILE_ERROR)
		return 0;
	
	OP_ASSERT (hfile);
	
	return (OPaC_FileDesc)(hfile);
}


OPaC_FileDesc
OPaC_OpenFileAppend (OPStringIn name)
{
	OP_FUNCTION (OPaC_OpenFileAppend, 1);
	OP_ARG (1, OPStringIn, name);
	
	HFILE    hfile;
	OFSTRUCT info;
	UINT     mode = OF_WRITE;
	
	char buffer[250];
	name->GetStringCopy (buffer, 250);
	
	hfile = OpenFile (buffer, &info, mode);
	
	if (hfile == HFILE_ERROR)
		return 0;
	
	OP_ASSERT (hfile);
	
	_llseek (hfile, 0L, 2);
	
	return (OPaC_FileDesc)(hfile);
}


Size
OPaC_ReadFile (OPaC_FileDesc file, void* void_buffer, Size size)
{
	OP_FUNCTION (OPaC_ReadFile, 3);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ARG (2, void*, void_buffer);
	OP_ARG (3, Size, size);
	
	HFILE hfile = (HFILE)(file);
	OP_ASSERT (hfile);
	
	Size       total_read = 0;
	UINT       chunk_size = 0;
	LatinChar* buffer     = (LatinChar*)(void_buffer);
	
	while (size > 10000) {
		
		chunk_size = _lread (hfile, buffer, 10000);
		total_read = total_read + chunk_size;
		buffer     = buffer + chunk_size;
		size       = size - 10000;
		
		if (chunk_size != 10000) return total_read;
	}
	
	chunk_size = _lread (hfile, buffer, (UINT)(size));
	total_read = total_read + chunk_size;
	
	return total_read;
}


Size
OPaC_WriteFile (OPaC_FileDesc file, const void* void_buffer, Size size)
{
	OP_FUNCTION (OPaC_WriteFile, 3);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ARG (2, const void*, void_buffer);
	OP_ARG (3, Size, size);
	
	HFILE hfile = (HFILE)(file);
	OP_ASSERT (hfile);
	
	Size             total_write = 0;
	UINT             chunk_size  = 0;
	const LatinChar* buffer      = (const LatinChar*)(void_buffer);
	
	while (size > 10000) {
		
		chunk_size  = _lwrite (hfile, buffer, 10000);
		total_write = total_write + chunk_size;
		buffer      = buffer + chunk_size;
		size        = size - 10000;
		
		if (chunk_size != 10000) return total_write;
	}
	
	chunk_size  = _lwrite (hfile, buffer, (UINT)(size));
	total_write = total_write + chunk_size;
	
	return total_write;
}


Bool
OPaC_SeekFileAbs (OPaC_FileDesc file, Int32 offset)
{
	OP_FUNCTION (OPaC_SeekFileAbs, 2);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ARG (2, Int32, offset);
	
	HFILE hfile = (HFILE)(file);
	UINT  mode  = (offset < 0) ? 2 : 0;
	DWORD pos   = (offset < 0) ? -offset - 1 : offset;
	
	OP_ASSERT (hfile);
	
	return (_llseek (hfile, pos, mode) == HFILE_ERROR) ? FALSE : TRUE;
}


Bool
OPaC_SeekFileRel (OPaC_FileDesc file, Int32 offset)
{
	OP_FUNCTION (OPaC_SeekFileRel, 2);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ARG (2, Int32, offset);
	
	HFILE hfile = (HFILE)(file);
	UINT  mode  = 1;
	DWORD pos   = offset;
	
	OP_ASSERT (hfile);
	
	return (_llseek (hfile, pos, mode) == HFILE_ERROR) ? FALSE : TRUE;
}


Size
OPaC_TellFilePos (OPaC_FileDesc file)
{
	OP_FUNCTION (OPaC_TellFilePos, 1);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ASSERT (file);
	
	HFILE hfile = (HFILE)(file);
	UINT  mode  = 1;
	DWORD pos   = 0;

	LONG offset = _llseek (hfile, pos, mode);
	
	OP_ASSERT (offset >= 0);
	
	return (Size)(offset);
}


Size
OPaC_TellFileSize (OPaC_FileDesc file)
{
	OP_FUNCTION (OPaC_TellFileSize, 1);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ASSERT (file);
	
	HFILE hfile = (HFILE)(file);
	LONG  size  = _filelength (hfile);
	
	return (size >= 0) ? (Size)(size) : 0;
}

Bool
OPaC_FlushFile (OPaC_FileDesc file)
{
	OP_FUNCTION (OPaC_FlushFile, 1);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ASSERT (file);
	
	return TRUE;
}



Bool
OPaC_TruncFile (OPaC_FileDesc file)
{
	OP_FUNCTION (OPaC_TruncFile, 1);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ASSERT (file);
	
	HFILE hfile = (HFILE)(file);
	UINT  mode  = 1;
	DWORD pos   = 0;
	
	LONG size = _llseek (hfile, pos, mode);
	
	if (size < 0) {
		return FALSE;
	}
	
	return (_chsize (hfile, size)) >= 0 ? TRUE : FALSE;
}



Bool
OPaC_ResizeFile (OPaC_FileDesc file, Size size)
{
	OP_FUNCTION (OPaC_ResizeFile, 2);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ARG (2, Size, size);
	OP_ASSERT (file);
	
	HFILE hfile = (HFILE)(file);
	
	return (_chsize (hfile, 0)) >= 0 ? TRUE : FALSE;
}



Bool
OPaC_CloseFile (OPaC_FileDesc file)
{
	OP_FUNCTION (OPaC_CloseFile, 1);
	OP_ARG (1, OPaC_FileDesc, file);
	OP_ASSERT (file);
	
	HFILE hfile = (HFILE)(file);
	return (_lclose (hfile) == HFILE_ERROR) ? FALSE : TRUE;
}

