/*
 *	bitsetinfo.cxx
 *
 *	This file defines a bitset description class, which is used
 *	to make bitsets more user-friendly than simple numbers.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/bitsetinfo.h>
#include <OPaC/new.h>
#include <OPaC/opacirt.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPBitsetInfo);
OP_IMPLEMENT_TYPEENUM (OPBitsetDataType);

/****************************************************************/

struct OPBitsetInfo::BitsetMember
{
	OPBitsetDataType	type;
	Card32				value;
	OPStringRef			name;
	
	BitsetMember () : type (OP_BITSET_DATA_MASK), value (0) { }
	virtual ~ BitsetMember () { }
};

struct OPBitsetInfo::BitsetGroup
{
	Count				num;
	BitsetMember**		members;
};

/****************************************************************/

OPBitsetInfo::OPBitsetInfo ()
{
	OP_METHOD (OPBitsetInfo, 0);
	
	this->group_array = 0;
	this->group_num   = 0;
}

OPBitsetInfo::~OPBitsetInfo ()
{
	OP_METHOD (~OPBitsetInfo, 0);
	
	if (this->group_array) {
		
		for (Count group = 0; group < this->group_num; group++) {
			for (Count member = 0; member < this->group_array[group].num; member++) {
				DELETE_FROM_ZONE (OPaC_sys_zone, this->group_array[group].members[member]);
			}
			OP_FREE_FROM_ZONE (OPaC_sys_zone, this->group_array[group].members);
		}
		
		OP_FREE_FROM_ZONE (OPaC_sys_zone, this->group_array);
		
		this->group_array = 0;
		this->group_num   = 0;
	}
}

void
OPBitsetInfo::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->group_num;
	storage << OPStorageSync::mark;
	
	for (Count group = 0; group < this->group_num; group++) {

		storage << this->group_array[group].num;
		storage << OPStorageSync::mark;
		
		for (Count member = 0; member < this->group_array[group].num; member++) {
			
			Card8 type = (Card8) this->group_array[group].members[member]->type;
			
			storage << type;
			storage << this->group_array[group].members[member]->value;
			storage << this->group_array[group].members[member]->name;
			storage << OPStorageSync::mark;
		}
	}
}

void
OPBitsetInfo::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	Count num = 0;
	
	storage >> num;
	storage >> OPStorageSync::mark;
	
	this->group_num   = num;
	this->group_array = num ? OP_ALLOC_ARRAY_FROM_ZONE (OPaC_sys_zone, BitsetGroup, num) : 0;
	
	for (Count group = 0; group < num; group++) {

		Count num = 0;
		
		storage >> num;
		storage >> OPStorageSync::mark;
		
		this->group_array[group].num = num;
		this->group_array[group].members = OP_ALLOC_ARRAY_FROM_ZONE (OPaC_sys_zone, BitsetMember*, num);
		
		for (Count member = 0; member < this->group_array[group].num; member++) {
			
			this->group_array[group].members[member] = NEW_FROM_ZONE (OPaC_sys_zone) BitsetMember;
			
			Card8 type;
			
			storage >> type;
			storage >> this->group_array[group].members[member]->value;
			storage >> this->group_array[group].members[member]->name;
			storage >> OPStorageSync::mark;
			
			this->group_array[group].members[member]->type = (OPBitsetDataType) type;
		}
	}
}


/*
 *	Initialise the bitset information from data collected
 *	thanks to the run-time data publishing class.
 */

void
OPBitsetInfo::Initialise (const OPaC_PubData* info)
{
	OP_METHOD (Initialise, 1);
	OP_ARG (1, const OPaC_PubData*, info);
	
	Count num = info->bitset_num;
	const OPaC_PubData::BitsetData* iter = info->bitset_list;
	
	//	A bitset may consist of several distinct groups of bits.
	
	this->group_num   = num;
	this->group_array = num ? OP_ALLOC_ARRAY_FROM_ZONE (OPaC_sys_zone, BitsetGroup, num) : 0;
	
	//	For each group, there is a list of members, which describe either the
	//	mask, the values or the bits of the bitset.
	
	for (Count group = 0; group < num; group++) {
		
		this->group_array[group].num = iter->num;
		this->group_array[group].members = OP_ALLOC_ARRAY_FROM_ZONE (OPaC_sys_zone, BitsetMember*, iter->num);
		
		for (Count member = 0; member < iter->num; member++) {
			this->group_array[group].members[member]= NEW_FROM_ZONE (OPaC_sys_zone) BitsetMember;
			this->group_array[group].members[member]->type  = iter->def[member].type;
			this->group_array[group].members[member]->value = iter->def[member].value;
			this->group_array[group].members[member]->name  = iter->def[member].name;
		}
		
		iter = iter->next;
	}
}


/*
 *	Return the number of groups defining the bitset.
 */

Count
OPBitsetInfo::CountGroups () const
{
	OP_METHOD (CountGroups, 0);
	
	return this->group_num;
}


/*
 *	Return the number of members in the specified group.
 */

Count
OPBitsetInfo::CountGroupMembers (Count group) const
{
	OP_METHOD (CountGroupMembers, 1);
	OP_ARG (1, Count, group);
	
	return (group < this->group_num) ? this->group_array[group].num : 0;
}


/*
 *	Return the description of the specified member/group. The
 *	type is one of the following values :
 *
 *	- OP_BITSET_DATA_MASK,  value is a mask
 *	- OP_BITSET_DATA_BIT,   value is a single bit
 *	- OP_BITSET_DATA_VALUE, value is an element of an enumeration
 */

Bool
OPBitsetInfo::GetDescription (Count group, Count member,
							  OPBitsetDataTypeOut type, Card32Out value, OPStringOut name)
{
	OP_METHOD (GetDescription, 5);
	OP_ARG (1, Count, group);
	OP_ARG (2, Count, member);
	OP_ARG (3, OPBitsetDataTypeOut, type);
	OP_ARG (4, Card32Out, value);
	OP_ARG (5, OPAtomOut, name);
	
	if ( (group < this->group_num)
	  && (member < this->group_array[group].num) ) {
		type  = this->group_array[group].members[member]->type;
		value = this->group_array[group].members[member]->value;
		name  = this->group_array[group].members[member]->name;
		return TRUE;
	}
	
	return FALSE;
}
