/*
 *	cell.cxx
 *
 *	Cell class implementation.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/cell.h>
#include <OPaC/except.h>
#include <OPaC/lookman.h>
#include <OPaC/grail.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENTATION (cell.h);
OP_IMPLEMENT_CLASS (OPCell);

/****************************************************************/

/*
 *	Initialise an empty cell and define its behaviour
 *	according to the Look Manager's policy.
 */

OPCell::OPCell ()
{
	OP_METHOD (OPCell, 0);
	
	this->shape |= OP_CTRL_AUTOHILITE;
	this->shape |= OP_CTRL_AUTOACTHILITE;
	this->shape |= OP_CTRL_BROADHILITE;
	this->shape |= OP_CTRL_NOFRAME;
	
	this->do_press_on_action  = OPLookMan::Current ()->IsPolicyHiliteCell () == FALSE;
	this->do_hilite_on_action = OPLookMan::Current ()->IsPolicyHiliteCell () == TRUE;
	
	int index;
	
	for (index = 0; index < OP_CELL_CTRL; index++) {
		this->control[index]  = 0;
		this->align[index]    = OP_CELL_ALIGN_LEFT;
		this->space[0][index] = 0;
		this->space[1][index] = 0;
	}
	
	this->border[OP_CELL_BORDER_TOP]    = FALSE;
	this->border[OP_CELL_BORDER_BOTTOM] = FALSE;
	this->border[OP_CELL_BORDER_LEFT]   = FALSE;
	this->border[OP_CELL_BORDER_RIGHT]  = FALSE;
}



/*
 *	The controls which have been inserted in the cell will
 *	be deleted with the cell itself, if they are no longer
 *	needed elsewhere.
 */

OPCell::~OPCell ()
{
	OP_METHOD (~OPCell, 0);
}



/*
 *	Save the cell, its associated controls and the frame
 *	definition.
 */

void
OPCell::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	int index;
	
	for (index = 0; index < OP_CELL_CTRL; index++) {
		storage << this->control[index];
		storage << this->align[index];
		storage << this->space[0][index] << this->space[1][index];
		storage << OPStorageSync::mark;
	}
	
	storage << this->border[OP_CELL_BORDER_TOP];
	storage << this->border[OP_CELL_BORDER_BOTTOM];
	storage << this->border[OP_CELL_BORDER_LEFT];
	storage << this->border[OP_CELL_BORDER_RIGHT];
	storage << OPStorageSync::mark;
}



/*
 *	Restore the cell, its associated controls and the frame
 *	definition.
 */

void
OPCell::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	int index;
	
	for (index = 0; index < OP_CELL_CTRL; index++) {
		storage >> this->control[index];
		storage >> this->align[index];
		storage >> this->space[0][index] >> this->space[1][index];
		storage >> OPStorageSync::mark;
	}
	
	storage >> this->border[OP_CELL_BORDER_TOP];
	storage >> this->border[OP_CELL_BORDER_BOTTOM];
	storage >> this->border[OP_CELL_BORDER_LEFT];
	storage >> this->border[OP_CELL_BORDER_RIGHT];
	storage >> OPStorageSync::mark;
}



/*
 *	Update the geometry of the controls contained by the
 *	cell.
 */

void
OPCell::UpdateGeometry ()
{
	OP_METHOD (UpdateGeometry, 0);
	
	OPCoord x1 = this->box.X1 ();
	OPCoord y1 = this->box.Y1 ();
	OPCoord x2 = this->box.X2 ();
	OPCoord y2 = this->box.Y2 ();
	OPCoord wf = (this->shape & OP_CTRL_NOFRAME) ? 0 : (OPLookMan::Current ()->ReturnWidgetFrameSize () + 1);
	OPCoord dx = (this->box.DX ()) - 2 * wf;
	OPCoord dy = (this->box.DY ()) - 2 * wf;
	
	OPCoord left   = 0;
	OPCoord center = 0;
	OPCoord right  = 0;
	
	Count index;
	
	for (index = 0; index < OP_CELL_CTRL; index++) {
		if (this->control[index].IsValid ()) {
			
			OPCoord ddx = 0;
			OPCoord ddy = 0;
			
			this->control[index]->GetSize (ddx, ddy);
			
			ddx += this->space[0][index];
			ddx += this->space[1][index];
			
			switch (this->align[index]) {
				
				case OP_CELL_ALIGN_LEFT:
					left   += ddx;
					break;
				
				case OP_CELL_ALIGN_CENTER:
					center += ddx;
					break;
				
				case OP_CELL_ALIGN_RIGHT:
					right  += ddx;
					break;
			}
		}
	}
	
	if (this->is_being_pressed) {
		x1++;
		y1++;
	}
	
	x1 += wf;
	y1 += wf;
	x2 -= wf;
	y2 -= wf;
	
	OPCoord pos_left   = x1;
	OPCoord pos_right  = x2 - right + 1;
	OPCoord pos_center = x1 + left + (dx - left - right - center) / 2;
	
	
	//	Now we know exactly how much place the different elements
	//	take up. We must now update their position...
	
	for (index = 0; index < OP_CELL_CTRL; index++) {
		if (this->control[index].IsValid ()) {
			
			OPCoord ddx = 0;
			OPCoord ddy = 0;
			
			this->control[index]->GetSize (ddx, ddy);
			
			ddx += this->space[0][index];
			ddx += this->space[1][index];
			
			switch (this->align[index]) {
				
				case OP_CELL_ALIGN_LEFT:
					this->control[index]->SetOrigin (pos_left + this->space[0][index], y1 + (dy - ddy)/2);
					pos_left += ddx;
					break;
				
				case OP_CELL_ALIGN_CENTER:
					this->control[index]->SetOrigin (pos_center + this->space[0][index], y1 + (dy - ddy)/2);
					pos_center += ddx;
					break;
				
				case OP_CELL_ALIGN_RIGHT:
					this->control[index]->SetOrigin (pos_right + this->space[0][index], y1 + (dy - ddy)/2);
					pos_right  += ddx;
					break;
			}
		}
	}
}



/*
 *	Called when the geometry changes.
 */

void
OPCell::ChangeGeometry (OPRectIn box)
{
	OP_METHOD (ChangeGeometry, 1);
	OP_ARG (1, OPRectIn, box);
	
	super::ChangeGeometry (box);
	this->UpdateGeometry ();
}



/*
 *	Prepare the cell for its destruction.
 */

void
OPCell::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->EmptyCell ();
	
	super::Delete ();
}



/*
 *	Cleanup the cell : no longer any control in it.
 *	This automatically disables the cell.
 */

Bool
OPCell::EmptyCell ()
{
	OP_METHOD (EmptyCell, 0);
	
	OPCellRef none  = 0;
	Count     index = OP_CELL_CTRL;
	
	while (index--) {
		if (this->control[index].IsValid ()) {
			this->control[index]->SetParent (none);
			this->control[index] = 0;
		}
	}

	this->SetGlobalEnable (FALSE);
	
	return TRUE;
}



/*
 *	Add a control to the cell. The controls are positioned
 *	automatically, depending on the specified alignment.
 *	Return FALSE if the index was out of range.
 */

Bool
OPCell::SetControl (Count index,
					OPControlIn control,
					OPCellAlignment align,
					OPCoord left_space,
					OPCoord right_space)
{
	OP_METHOD (SetControl, 5);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPControlIn, control);
	OP_ARG (3, OPCellAlignment, align);
	OP_ARG (4, OPCoord, left_space);
	OP_ARG (5, OPCoord, right_space);
	
	OPCellRef none = 0;
	
	if (index >= OP_CELL_CTRL) {
		return FALSE;
	}
	
	if (this->control[index].IsValid ()) {
		if (this->control[index] == control) {
			return TRUE;
		}
		
		this->control[index]->SetParent (none);
		this->control[index] = 0;
	}
	
	this->control[index]  = control;
	this->align[index]    = align;
	this->space[0][index] = left_space;
	this->space[1][index] = right_space;
	
	if (control.IsValid ()) {
		
		if (this->ReturnGlobalEnable () == FALSE) {
			this->SetGlobalEnable (TRUE);
		}
		
		control->SetParent (OPCellRef (this));
		control->SetFlagOkDetect (FALSE);
		control->SetFlagOkUIMove (FALSE);
		control->SetShapeStyle (this->shape);
		
		this->UpdateGeometry ();
	}
	
	return TRUE;
}



/*
 *	Same as "SetControl" with index, but this method uses
 *	the first available index... Return FALSE if there is
 *	no more room left in the cell.
 */

Bool
OPCell::SetControl (OPControlIn control,
					OPCellAlignment align,
					OPCoord left_space,
					OPCoord right_space)
{
	OP_METHOD (SetControl, 4);
	OP_ARG (1, OPControlIn, control);
	OP_ARG (2, OPCellAlignment, align);
	OP_ARG (3, OPCoord, left_space);
	OP_ARG (4, OPCoord, right_space);
	
	int index;
	
	for (index = 0; index < OP_CELL_CTRL; index++) {
		if (this->control[index].IsInvalid ()) {
			return this->SetControl (index, control, align, left_space, right_space);
		}
	}
	
	return FALSE;
}



/*
 *	Set the alignment of the specified element.
 */

Bool
OPCell::SetAlignment (Count index, OPCellAlignment alignment)
{
	OP_METHOD (SetAlignment, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPCellAlignment, alignment);
	
	if (index >= OP_CELL_CTRL) {
		return FALSE;
	}
	
	this->align[index] = alignment;
	
	return TRUE;
}



/*
 *	Set the space at the left of the specified element.
 */

Bool
OPCell::SetLeftSpace (Count index, OPCoord space)
{
	OP_METHOD (SetAlignment, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPCoord, space);
	
	if (index >= OP_CELL_CTRL) {
		return FALSE;
	}
	
	this->space[0][index] = space;
	
	return TRUE;
}



/*
 *	Set the space at the left of the specified element.
 */

Bool
OPCell::SetRightSpace (Count index, OPCoord space)
{
	OP_METHOD (SetAlignment, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPCoord, space);
	
	if (index >= OP_CELL_CTRL) {
		return FALSE;
	}
	
	this->space[1][index] = space;
	
	return TRUE;
}



/*
 *	Return the specified control (or 0 if out of range).
 */

Bool
OPCell::GetControl (Count index, OPControlOut control) const
{
	OP_METHOD (GetControl, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPControlOut, control);
	
	if (index < OP_CELL_CTRL) {
		control = this->control[index];
		return control.IsValid ();
	}
	
	return FALSE;
}



/*
 *	Look for the specified control in the cell and return
 *	TRUE if found; "index" will be updated accordingly.
 */

Bool
OPCell::FindControl (OPControlIn control, CountOut user_index)
{
	OP_METHOD (FindControl, 2);
	OP_ARG (1, OPControlIn, control);
	OP_ARG (2, CountOut, user_index);
	
	int index;
	
	for (index = 0; index < OP_CELL_CTRL; index++) {
		if (this->control[index] == control) {
			user_index = index;
			return TRUE;
		}
	}
	
	return FALSE;
}



/*
 *	Sets the cell border state (top, left, bottom and right
 *	borders).
 *
 *	Throws :
 *
 *	  - OPExcAccess (OUT_OF_RANGE) if index is out of range.
 */

void
OPCell::SetCellBorder (Count which, Bool state)
{
	OP_METHOD (SetCellBorder, 2);
	OP_ARG (1, Count, which);
	OP_ARG (2, Bool, state);
	
	if (which > 4) {
		 THROW (OPExcAccess (OPExcAccess::OUT_OF_RANGE));
	}
	
	this->border[which] = state;
}



/*
 *	Gets the cell border state (top, left, bottom and right
 *	borders).
 *
 *	Throws :
 *
 *	  - OPExcAccess (OUT_OF_RANGE) if index is out of range.
 */

Bool
OPCell::GetCellBorder (Count which) const
{
	OP_METHOD (GetCellBorder, 1);
	OP_ARG (1, Count, which);
	
	if (which > 4) {
		 THROW (OPExcAccess (OPExcAccess::OUT_OF_RANGE));
	}
	
	return this->border[which];
}




/*
 *	Return the cell's minimal size. This is equal to the
 *	horizontal sum of minima and the vertical maximum
 *	minima.
 */

void
OPCell::GetMinSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	OPCoord dx1 = 0;
	OPCoord dy1 = 0;
	OPCoord dx2 = 0;
	OPCoord dy2 = 0;
	
	Count index;
	
	for (index = 0; index < OP_CELL_CTRL; index++) {
		if (this->control[index].IsValid ()) {
			this->control[index]->GetMinSize (dx1, dy1);
			dx1 += this->space[0][index];
			dx1 += this->space[1][index];
			dx2 += dx1;
			dy2  = (dy2 > dy1) ? dy2 : dy1;
		}
	}
	
	OPCoord wf = (this->shape & OP_CTRL_NOFRAME) ? 0 : (OPLookMan::Current ()->ReturnWidgetFrameSize () + 1);
	
	dx = dx2 + 2 * wf;
	dy = dy2 + 2 * wf;
}



/*
 *	Return the cell's maximum size. Return an arbitrary
 *	huge value.
 */

void
OPCell::GetMaxSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMaxSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	super::GetMaxSize (dx, dy);
}



/*
 *	The only message a cell understands is that the look
 *	manager has changed.
 */

void
OPCell::Broadcast (OPAtomIn message, OPObjectIn object)
{
	OP_METHOD (Broadcast, 2);
	OP_ARG (1, OPAtomIn, message);
	OP_ARG (2, OPObjectIn, object);
	
	if (message == this->atom_look_man_changed) {
		this->do_press_on_action  = OPLookMan::Current ()->IsPolicyHiliteCell () == FALSE;
		this->do_hilite_on_action = OPLookMan::Current ()->IsPolicyHiliteCell () == TRUE;
		this->UpdateGeometry ();
	}
	
	super::Broadcast (message, object);
}



/*
 *	Draw the cell. This will draw the cell background, the
 *	frame border and then let the view draw the included
 *	controls.
 */

void
OPCell::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	super::Draw (port, rect, what);
	
	if (what & OP_DRAW_OBJECT) {
		
		const OPAdornerRef& adorner = port->ReturnAdorner ();
		
		OPLookFrame frame  = 0;
		OPCoord     margin = (this->shape & OP_CTRL_NOFRAME) ? 0 : (OPLookMan::Current ()->ReturnWidgetFrameSize () + 1);
		
		//	Draw the 4 edges of the frame... or less.
		
		if (this->border[OP_CELL_BORDER_TOP]) {
			frame |= OP_FRAME_TOP;
		}
		
		if (this->border[OP_CELL_BORDER_BOTTOM]) {
			frame |= OP_FRAME_BOTTOM;
		}
		
		if (this->border[OP_CELL_BORDER_LEFT]) {
			frame |= OP_FRAME_LEFT;
		}
		
		if (this->border[OP_CELL_BORDER_RIGHT]) {
			frame |= OP_FRAME_RIGHT;
		}
		
		if ((this->shape & OP_CTRL_NOFRAME) == 0) {
			adorner->DrawFrame (this->box, frame, margin);
		}
	}
}

