/*
 *	cellgrid.cxx
 *
 *	The Cell Grid class manages a collection of cells, which can be arranged
 *	in a two dimensional array.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/array.h>
#include <OPaC/cellgrid.h>
#include <OPaC/control.h>
#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/except.h>

/****************************************************************/

OP_IMPLEMENTATION (cellgrid.h);
OP_IMPLEMENT_CLASS (OPCellGrid);
	
/****************************************************************/


/*
 *	Initialise an empty cell grid.
 */

OPCellGrid::OPCellGrid ()
{
	OP_METHOD (OPCellGrid, 0);
	
	this->cell_dx = 0;
	this->cell_dy = 0;
	
	this->is_isometric = TRUE;
	
	this->cell_dx_array = 0;
	this->cell_dy_array = 0;
	this->cell_ox_array = 0;
	this->cell_oy_array = 0;
	
	this->offset_x = 0;
	this->offset_y = 0;
	
	this->margin          = 0;
	this->margin_x_left   = 0;
	this->margin_x_right  = 0;
	this->margin_y_top    = 0;
	this->margin_y_bottom = 0;
	
	this->min_cell_dx = 4;
	this->min_cell_dy = 4;
	this->max_cell_dx = 20000;
	this->max_cell_dy = 20000;
	
	this->h_cell_count = 0;
	this->v_cell_count = 0;
	
	this->is_geom_ok   = FALSE;
	this->is_minmax_ok = FALSE;
	
	this->draw_frame_high = TRUE;
	this->draw_frame_low  = FALSE;
	
	this->draw_cache   = 0;
	this->draw_h_cache = 0;
	this->draw_v_cache = 0;
	
	//	The ghost cell is a non-empty cell which may be neither drawn
	//	nor detected... It will be used when a cell overlaps several
	//	lines or columns.
	
	this->ghost_cell = OP_NEW (OPControl);
	this->ghost_cell->SetSize (0, 0);
}


void
OPCellGrid::SetDrawFrameHigh (Bool yes)
{
	OP_METHOD (SetDrawFrameHigh, 1);
	OP_ARG (1, Bool, yes);
	
	this->draw_frame_high = yes;
	this->draw_frame_low  = FALSE;
}

void
OPCellGrid::SetDrawFrameLow (Bool yes)
{
	OP_METHOD (SetDrawFrameLow, 1);
	OP_ARG (1, Bool, yes);
	
	this->draw_frame_high = FALSE;
	this->draw_frame_low  = yes;
}



/*
 *	Free a cell grid (and its associated cells).
 */

OPCellGrid::~OPCellGrid ()
{
	OP_METHOD (~OPCellGrid, 0);
	
	if ( (this->array.IsValid ())
	  && (this->array->TotalCount () != 0) ) {
		this->SetArraySize (0, 0);
	}
	
	this->DiscardSizeInfo ();
}


void
OPCellGrid::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->cell_dx << this->cell_dy << this->margin << this->is_isometric;
	storage << this->h_cell_count << this->v_cell_count;
	storage << this->ghost_cell;
	storage << this->array;
	storage << this->draw_frame_high << this->draw_frame_low;
	storage << OPStorageSync::mark;
}


void
OPCellGrid::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	OP_ASSERT ( (this->array.IsInvalid ())
			 || (this->array->TotalCount () == 0) );
	
	super::RestoreInstance (storage);
	
	storage >> this->cell_dx >> this->cell_dy >> this->margin >> this->is_isometric;
	storage >> this->h_cell_count >> this->v_cell_count;
	storage >> this->ghost_cell;
	storage >> this->array;
	storage >> this->draw_frame_high >> this->draw_frame_low;
	storage >> OPStorageSync::mark;
	
	this->is_geom_ok   = FALSE;
	this->is_minmax_ok = FALSE;
}



void
OPCellGrid::NotifyArraySize (Count dx, Count dy)
{
	OP_METHOD (NotifyArraySize, 2);
	OP_ARG (1, Count, dx);
	OP_ARG (2, Count, dy);
}

void
OPCellGrid::NotifyAttachCell (OPControlIn cell, Count x, Count y)
{
	OP_METHOD (NotifyAttachCell, 3);
	OP_ARG (1, OPControlIn, cell);
	OP_ARG (2, Count, x);
	OP_ARG (3, Count, y);
}

void
OPCellGrid::NotifyDetachCell (OPControlIn cell, Count x, Count y)
{
	OP_METHOD (NotifyDetachCell, 3);
	OP_ARG (1, OPControlIn, cell);
	OP_ARG (2, Count, x);
	OP_ARG (3, Count, y);
}



/*
 *	Prepare the cell grid for its deletion. This will remove
 *	every cell from the grid.
 */

void
OPCellGrid::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->SetArraySize (0, 0);
	
	super::Delete ();
}



/*
 *	Return the specified cell's dimensions. This is easy as long
 *	as each cell has the same size. Otherwise, rely on the cell's
 *	own dimensions.
 */

void
OPCellGrid::GetCellRect (Count h, Count v, OPRectOut rect) const
{
	OP_METHOD (GetCellRect, 3);
	OP_ARG (1, Count, h);
	OP_ARG (2, Count, v);
	OP_ARG (3, OPRectOut, rect);
	
	if (this->is_isometric) {
		
		//	Each cell has the same size. Find out how many (integer) pixels each
		//	cell occupies, then compute is position accordingly.
		
		Int32 cdx = (Int32)(this->box.DX () - this->margin_x_left - this->margin_x_right);
		Int32 cdy = (Int32)(this->box.DY () - this->margin_y_top  - this->margin_y_bottom);
		
		cdx /= this->h_cell_count;
		cdy /= this->v_cell_count;
		
		//	Constrain to the maximum cell size, in order to avoid having cells which
		//	grow outside the cell grid.
		
		Int32 ldx = cdx > this->cell_dx ? (Int32)(this->cell_dx) : cdx;
		Int32 ldy = cdy > this->cell_dy ? (Int32)(this->cell_dy) : cdy;
		
		OPCoord x1 = h * this->cell_dx + this->margin_x_left + this->offset_x;
		OPCoord y1 = v * this->cell_dy + this->margin_y_top  + this->offset_y;
		OPCoord x2 = x1 + ldx - 1;
		OPCoord y2 = y1 + ldy - 1;
		
		x1 += this->box.X1 ();
		x2 += this->box.X1 ();
		y1 += this->box.Y1 ();
		y2 += this->box.Y1 ();
		
		rect = OPRect (x1, y1, x2, y2);
		
	} else {
		
		//	If the cell are not each of the same size, we rely on the computed
		//	size and position, which are initialised by the UpdateCellGeometry
		//	method. Beware not to use GetCellRect from UpdateCellGeometry if the
		//	cells are not isometric.
		
		if ( (! this->is_geom_ok)
		  || (this->cell_dx_array == 0)
		  || (this->cell_dy_array == 0)
		  || (this->cell_ox_array == 0)
		  || (this->cell_oy_array == 0) ) {
			
			this->UpdateCellGeometry ();
		}
		
		OPCoord x1 = this->box.X1 () + this->margin_x_left + this->offset_x;
		OPCoord y1 = this->box.Y1 () + this->margin_y_top  + this->offset_y;
		
		x1 += this->cell_ox_array[h];
		y1 += this->cell_oy_array[v];
		
		OPCoord x2 = x1 + this->cell_dx_array[h] - 1;
		OPCoord y2 = y1 + this->cell_dy_array[v] - 1;
		
		rect = OPRect (x1, y1, x2, y2);
	}
}

/*
 *	Return the minimum size into which the specified cell
 *	would fit.
 */

void
OPCellGrid::GetCellMinSize (Count h, Count v, OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetCellMinSize, 4);
	OP_ARG (1, Count, h);
	OP_ARG (2, Count, v);
	OP_ARG (3, OPCoordOut, dx);
	OP_ARG (4, OPCoordOut, dy);
	
	OPControlRef control;
	
	if (this->GetCell (h, v, control)) {
		control->GetMinSize (dx, dy);
	} else {
		dx = this->cell_dx;
		dy = this->cell_dy;
	}
}


/*
 *	Compute new cell minimum and maximum dimensions. This
 *	will do the calculation only if the min/max sizes are
 *	known to have changed.
 *
 *	If the minimum size happens to be smaller than the
 *	current cell size, then the cell size will be changed.
 */

void
OPCellGrid::ComputeCellMinMax ()
{
	OP_METHOD (ComputeCellMinMax, 0);
	
	if (this->is_minmax_ok)
		return;
	
	Count index = 0;
	Count num   = this->h_cell_count * this->v_cell_count;
	
	this->min_cell_dx = 2;
	this->min_cell_dy = 2;
	this->max_cell_dx = 1000;
	this->max_cell_dy = 1000;
	
	while (num) {
		if (this->array[index].IsValid ()) {
			
			OPCoord min_x = 0;
			OPCoord min_y = 0;
			OPCoord max_x = 0;
			OPCoord max_y = 0;
			
			this->array[index]->GetMinSize (min_x, min_y);
			this->array[index]->GetMaxSize (max_x, max_y);
			
			if (min_x > this->min_cell_dx) this->min_cell_dx = min_x;
			if (min_y > this->min_cell_dy) this->min_cell_dy = min_y;
			if (max_x < this->max_cell_dx) this->max_cell_dx = max_x;
			if (max_y < this->max_cell_dy) this->max_cell_dy = max_y;
		}
		
		index++;
		num--;
	}
	
	this->is_minmax_ok = TRUE;
	
	//	We might need to resize the cell...
	
	this->SetCellSize (this->cell_dx, this->cell_dy);
}


/*
 *	Compute the new margins depending on the frame and on the
 *	user defined margin.
 */

void
OPCellGrid::ComputeMargins ()
{
	OP_METHOD (ComputeMargins, 0);
	
	Bool    frame  = this->draw_frame_high || this->draw_frame_low;
	OPCoord effect = (frame) ? OPLookMan::Current ()->Return3DEffectWidth () : 0;
	
	//	The margins are currently equal to the 3D effect (if any) of
	//	the frame. The margins could be used for anything else.
	
	this->margin_x_left = this->margin_x_right  = effect + this->margin;
	this->margin_y_top  = this->margin_y_bottom = effect + this->margin;
}


/*
 *	Compute the new size for the cell grid, so that each cell
 *	will fit exactly into the grid.
 */

void
OPCellGrid::UpdateGridGeometry ()
{
	OP_METHOD (UpdateGridGeometry, 0);
	
	this->ComputeMargins ();
	
	OPCoord dx = 0;
	OPCoord dy = 0;
	
	dx += this->margin_x_left;
	dx += this->margin_x_right;
	dy += this->margin_y_top;
	dy += this->margin_y_bottom;
	
	if (this->is_isometric) {
		
		//	Every cell has the same size, so the total size is easy to
		//	compute.
		
		dx += this->cell_dx * this->h_cell_count;
		dy += this->cell_dy * this->v_cell_count;
		
	} else {
		
		//	Each cell might have a different size, so we must add the
		//	individual cell sizes.
		
		OPCoord cdx;
		OPCoord cdy;
		
		for (Count h = 0; h < this->h_cell_count; h++) {
			this->GetCellSize (h, 0, cdx, cdy);
			dx += cdx;
		}
		
		for (Count v = 0; v < this->v_cell_count; v++) {
			this->GetCellSize (0, v, cdx, cdy);
			dy += cdy;
		}
	}
	
	this->SetSize (dx, dy);
}


/*
 *	Insert a cell into the specified column. The insertion point
 *	is where the first cell is empty.
 */

void
OPCellGrid::InsertCellInColumn (OPControlIn cell, Count h)
{
	OP_METHOD (InsertCellInColumn, 2);
	OP_ARG (1, OPControlIn, cell);
	OP_ARG (2, Count, h);
	
	Count v = 0;
	OPControlRef test;
	
	//	Find first empty cell.
	
	while (v < this->v_cell_count) {
		if (this->GetCell (h, v, test) == FALSE) {
			this->SetCell (h, v, cell);
			return;
		}
		v++;
	}
	
	//	Grow the array so that there is enough room for the cell.
	
	this->SetArrayHeight (v+1);
	this->SetCell (h, v, cell);
}

void
OPCellGrid::InsertCellInLine (OPControlIn cell, Count v)
{
	OP_METHOD (InsertCellInLine, 2);
	OP_ARG (1, OPControlIn, cell);
	OP_ARG (2, Count, v);
	
	Count h = 0;
	OPControlRef test;
	
	//	Find first empty cell.
	
	while (h < this->h_cell_count) {
		if (this->GetCell (h, v, test) == FALSE) {
			this->SetCell (h, v, cell);
			return;
		}
		h++;
	}
	
	//	Grow the array so that there is enough room for the cell.
	
	this->SetArrayWidth (h+1);
	this->SetCell (h, v, cell);
}


/*
 *	Define the margin between the cells and the frame.
 */

void
OPCellGrid::SetMargin (OPCoord margin)
{
	OP_METHOD (SetMargin, 1);
	OP_ARG (1, OPCoord, margin);
	
	this->margin = margin;
	this->is_geom_ok = FALSE;
}


/*
 *	Return the margin.
 */

OPCoord
OPCellGrid::ReturnMargin () const
{
	OP_METHOD (ReturnMargin, 0);
	return this->margin;
}


/*
 *	Define the isometric mode. If the isometric mode is off,
 *	each cell will have its own - minimal - size.
 */

void
OPCellGrid::SetIsometricMode (Bool yes)
{
	OP_METHOD (SetIsometricMode, 1);
	OP_ARG (1, Bool, yes);
	
	if (this->is_isometric != yes) {
		this->is_isometric = yes;
		this->is_geom_ok   = FALSE;
	}
}

/*
 *	Return the isometric mode. TRUE means that each cell has
 *	the same size.
 */

Bool
OPCellGrid::ReturnIsometricMode () const
{
	OP_METHOD (ReturnIsometricMode, 0);
	
	return this->is_isometric;
}

/*
 *	Free the memory allocated for the size information.
 */

void
OPCellGrid::DiscardSizeInfo ()
{
	OP_METHOD (DiscardSizeInfo, 0);
	
	if (this->cell_dx_array) {
		OP_FREE (this->cell_dx_array);
		this->cell_dx_array = 0;
		this->is_geom_ok &= this->is_isometric;
	}
	if (this->cell_dy_array) {
		OP_FREE (this->cell_dy_array);
		this->cell_dy_array = 0;
		this->is_geom_ok &= this->is_isometric;
	}
	if (this->cell_ox_array) {
		OP_FREE (this->cell_ox_array);
		this->cell_ox_array = 0;
		this->is_geom_ok &= this->is_isometric;
	}
	if (this->cell_oy_array) {
		OP_FREE (this->cell_oy_array);
		this->cell_oy_array = 0;
		this->is_geom_ok &= this->is_isometric;
	}
}


/*
 *	Update geometry of every cell in the cell grid. This
 *	is done only when the geometry is known to have
 *	changed.
 */

void
OPCellGrid::UpdateCellGeometry () const
{
	OP_METHOD (UpdateCellGeometry, 0);
	
	OPCellGrid* that = const_cast<OPCellGrid*> (this);
	
	if (this->is_geom_ok) {
		return;
	}
	
	//	Geometry is not valid. Make sure the minimum/maximum size is
	//	correctly defined, then update the positions.
	
	that->ComputeCellMinMax ();
	that->ComputeMargins ();
	
	if (this->is_isometric) {
		
		//	Each cell has the same size. Therefore, it is easy to find out about
		//	the cell's position and size. First find out about the offset, so that
		//	GetCellRect will return proper values.
		
		OPCoord ldx = (that->box.DX () - that->margin_x_left - that->margin_x_right);
		OPCoord ldy = (that->box.DY () - that->margin_y_top  - that->margin_y_bottom);
		
		ldx /= that->h_cell_count;
		ldy /= that->v_cell_count;
		
		OPCoord ox  = (ldx - that->cell_dx) * that->h_cell_count / 2;
		OPCoord oy  = (ldy - that->cell_dy) * that->v_cell_count / 2;
		
		if (ox < 0) ox = 0;
		if (oy < 0) oy = 0;
		
		that->offset_x = ox;
		that->offset_y = oy;
		
		if (that->cell_dx > ldx) that->cell_dx = ldx;
		if (that->cell_dy > ldy) that->cell_dy = ldy;
		
		OPRect rect;
		
		//	Set the position and size of each cell in the grid by using the
		//	GetCellRect method. This is possible only because the cells are
		//	isometric.
		
		for (Count x = 0; x < that->h_cell_count; x++) {
			for (Count y = 0; y < that->v_cell_count; y++) {
				
				OPControlRef cell;
				that->GetCell (x, y, cell);
				
				if (cell.IsValid ()) {
					that->GetCellRect (x, y, rect);
					cell->SetOrigin (rect.X1 (), rect.Y1 ());
					cell->SetSize (rect.DX (), rect.DY ());
				}
			}
		}
		
	} else {
		
		that->DiscardSizeInfo ();
		
		//	Allocate the buffers for the cell sizes (by column and by line)
		//	and positions.
		
		that->cell_dx_array = OP_ALLOC_ARRAY (OPCoord, that->h_cell_count);
		that->cell_dy_array = OP_ALLOC_ARRAY (OPCoord, that->v_cell_count);
		that->cell_ox_array = OP_ALLOC_ARRAY (OPCoord, that->h_cell_count+1);
		that->cell_oy_array = OP_ALLOC_ARRAY (OPCoord, that->v_cell_count+1);
		
		Count   x  = 0;
		Count   y  = 0;
		OPCoord dx = 0;
		OPCoord dy = 0;
		OPCoord dx_total = 0;
		OPCoord dy_total = 0;
		
		//	Compute the cell sizes and positions column by column first...
		
		for (x = 0; x < that->h_cell_count; x++) {
			
			that->cell_dx_array[x] = 0;
			that->cell_ox_array[x] = dx_total;
			
			for (y = 0; y < that->v_cell_count; y++) {
				that->GetCellMinSize (x, y, dx, dy);
				if (dx > that->cell_dx_array[x]) {
					that->cell_dx_array[x] = dx;
				}
			}
			
			dx_total += that->cell_dx_array[x];
		}
		
		that->cell_ox_array[x] = dx_total;
		
		//	...and then line by line.
		
		for (y = 0; y < that->v_cell_count; y++) {
			
			that->cell_dy_array[y] = 0;
			that->cell_oy_array[y] = dy_total;
			
			for (x = 0; x < that->h_cell_count; x++) {
				that->GetCellMinSize (x, y, dx, dy);
				if (dy > that->cell_dy_array[y]) {
					that->cell_dy_array[y] = dy;
				}
			}
			
			dy_total += that->cell_dy_array[y];
		}
		
		that->cell_oy_array[y] = dy_total;
		
		//	Compute the offset to the top left cell, so that the cells are
		//	centered in the grid.
		
		OPCoord dx_room = (that->box.DX () - that->margin_x_left - that->margin_x_right);
		OPCoord dy_room = (that->box.DY () - that->margin_y_top  - that->margin_y_bottom);
		
		OPCoord ox = (dx_room - dx_total) / 2;
		OPCoord oy = (dy_room - dy_total) / 2;
		
		if (ox < 0) ox = 0;
		if (oy < 0) oy = 0;
		
		that->offset_x = ox;
		that->offset_y = oy;
		
		//	Now, define the position and the size of each cell. We cannot use the
		//	GetCellRect method here, since it relies on a valid geometry when the
		//	cells are not isometric.
		
		ox = this->box.X1 () + this->margin_x_left + this->offset_x;
		oy = this->box.Y1 () + this->margin_y_top  + this->offset_y;

		OPControlRef cell;
		
		for (x = 0; x < that->h_cell_count; x++) {
			for (y = 0; y < that->v_cell_count; y++) {
				
				that->GetCell (x, y, cell);
				
				if (cell.IsValid ()) {
					cell->SetOrigin (ox + that->cell_ox_array[x], oy + that->cell_oy_array[y]);
					cell->SetSize (that->cell_dx_array[x], that->cell_dy_array[y]);
				}
			}
		}
	}
	
	that->is_geom_ok = TRUE;
}



/*
 *	Define the cell's array size (width and height). If
 *	the array shrinks, some cells may get lost...
 */

Bool
OPCellGrid::SetArraySize (Count w, Count h)
{
	OP_METHOD (SetArraySize, 2);
	OP_ARG (1, Count, w);
	OP_ARG (2, Count, h);
	
	if ( (w == this->h_cell_count)
	  && (h == this->v_cell_count) ) {
		return TRUE;
	}
	
	Count num = w * h;
	
	if (num == 0) {
		
		if ( (this->array.IsInvalid ())
		  || (this->array->TotalCount () == 0) ) {
			return TRUE;
		}
		
		Count x, y;
		Count index = 0;
		
		for (x = 0; x < this->h_cell_count; x++) {
			for (y = 0; y < this->v_cell_count; y++) {
				if (this->array[index].IsValid ()) {
					this->NotifyDetachCell (this->array[index], x, y);
					this->array[index]->SetParent (OPViewNull);
					this->array[index] = 0;
				}
				index++;
			}
		}
		
		this->array = OP_NEW (OPArray);
		
		this->NotifyArraySize (w, h);
		
		this->h_cell_count = w;
		this->v_cell_count = h;
		
		return TRUE;
	}
	
	//	Allocate the array with the specified size.
	
	OPControlArray array;
	
	array = OP_NEW (OPArray);
	
	array->SetMemZone (this->ReturnMemZone ());
	array->SetSize (num);
	
	//	Copy previous array to the new array...
	
	if ( (this->array.IsValid ())
	  && (this->array->TotalCount ())) {
		
		Count x, y;
		
		for (x = 0; x < w; x++) {
			for (y = 0; y < h; y++) {
				
				//	Copy the cells and detach them. They will be re-attached to the
				//	grid when the array has been completely copied.
				
				this->GetCell (x, y, array[x*h+y]);
				this->NotifyDetachCell (array[x*h+y], x, y);
			}
		}
	}
	
	//	Store new array size.
	
	this->array = array;
	
	this->NotifyArraySize (w, h);
	
	this->h_cell_count = w;
	this->v_cell_count = h;
	
	this->is_minmax_ok = FALSE;
	this->is_geom_ok   = FALSE;
	
	//	Re-attach the cells to the grid.
	
	for (Count x = 0; x < this->h_cell_count; x++) {
		for (Count y = 0; y < this->v_cell_count; y++) {
			this->NotifyAttachCell (this->array[x*h+y], x, y);
		}
	}
	
	return TRUE;
}


/*
 *	Just change the horizontal array size. If it shrinks,
 *	then some cells will be lost.
 */

Bool
OPCellGrid::SetArrayWidth (Count w)
{
	OP_METHOD (SetArrayWidth, 1);
	OP_ARG (1, Count, w);
	
	return this->SetArraySize (w, this->ReturnArrayHeight ());
}



/*
 *	Just change the vertical array size. If it shrinks,
 *	then some cells will be lost.
 */

Bool
OPCellGrid::SetArrayHeight (Count h)
{
	OP_METHOD (SetArrayHeight, 1);
	OP_ARG (1, Count, h);
	
	return this->SetArraySize (this->ReturnArrayWidth (), h);
}



/*
 *	Return the array's width and height (expressed in # of
 *	cells).
 */

void
OPCellGrid::GetArraySize (CountOut w, CountOut h) const
{
	OP_METHOD (GetArraySize, 2);
	OP_ARG (1, CountOut, w);
	OP_ARG (2, CountOut, h);
	
	w = this->h_cell_count;
	h = this->v_cell_count;
}

Count
OPCellGrid::ReturnArrayWidth () const
{
	OP_METHOD (ReturnArrayWidth, 0);
	return this->h_cell_count;
}

Count
OPCellGrid::ReturnArrayHeight () const
{
	OP_METHOD (ReturnArrayHeight, 0);
	return this->v_cell_count;
}



/*
 *	Define the cell's size. The size may be changed due to
 *	some minimum/maximum boundaries.
 *
 *	The geometry will be updated only when needed : when
 *	drawing is requested or when some detection is needed.
 */

void
OPCellGrid::SetCellSize (OPCoord dx, OPCoord dy)
{
	OP_METHOD (GetCellSize, 2);
	OP_ARG (1, OPCoord, dx);
	OP_ARG (2, OPCoord, dy);
	
	if (this->is_isometric) {
		
		//	If the cells all have the same size, verify that the new size is not
		//	getting too small for the largest cell to fit...
		
		this->ComputeCellMinMax ();
		
		if (dx < this->min_cell_dx) dx = this->min_cell_dx;
		if (dy < this->min_cell_dy) dy = this->min_cell_dy;
		if (dx > this->max_cell_dx) dx = this->max_cell_dx;
		if (dy > this->max_cell_dy) dy = this->max_cell_dy;
	}
	
	if ( (this->cell_dx == dx)
	  && (this->cell_dy == dy) ) {
		return;
	}
	
	this->cell_dx = dx;
	this->cell_dy = dy;
	
	this->is_geom_ok = FALSE;
}



/*
 *	Return the typical cell's size. The default CellGrid
 *	sets every cell to the same size...
 */

void
OPCellGrid::GetCellSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetCellSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	dx = this->cell_dx;
	dy = this->cell_dy;
}

/*
 *	Return the real size of the specified cell. If the cells
 *	are not isometric, then this might be different from the
 *	results returned by GetCellSize method above.
 */

void
OPCellGrid::GetCellSize (Count h, Count v, OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetCellSize, 4);
	OP_ARG (1, Count, h);
	OP_ARG (2, Count, v);
	OP_ARG (3, OPCoordOut, dx);
	OP_ARG (4, OPCoordOut, dy);
	
	dx = this->cell_dx;
	dy = this->cell_dy;
	
	if (this->is_isometric == FALSE) {
		
		OPRect rect;
		
		//	Update the geometry of the cells and find about the specified cell's
		//	rectangle. Each cell can have a dimensions.
		
		this->UpdateCellGeometry ();
		this->GetCellRect (h, v, rect);
		
		dx = rect.DX ();
		dy = rect.DY ();
	}
}

/*
 *	Return the origin of the specified cell.
 */

void
OPCellGrid::GetCellPos (Count h, Count v, OPCoordOut x, OPCoordOut y) const
{
	OP_METHOD (GetCellPos, 4);
	OP_ARG (1, Count, h);
	OP_ARG (2, Count, v);
	OP_ARG (3, OPCoordOut, x);
	OP_ARG (4, OPCoordOut, y);
	
	OPRect rect;
	
	this->GetCellRect (h, v, rect);
	
	x = rect.X1 ();
	y = rect.Y1 ();
}

/*
 *	Set a cell into the cell array. If the cell falls out
 *	of the array, an exception will be generated.
 *
 *	Throws :
 *
 *	  - OPExcAccess (OUT_OF_RANGE) if either horizontal or
 *		vertical coordinates are out of range.
 */

void
OPCellGrid::SetCell (Count h, Count v, OPControlIn cell)
{
	OP_METHOD (SetCell, 3);
	OP_ARG (1, Count, h);
	OP_ARG (2, Count, v);
	OP_ARG (3, OPControlIn, cell);
	
	if ( (h >= this->h_cell_count)
	  || (v >= this->v_cell_count) ) {
		THROW (OPExcAccess (OPExcAccess::OUT_OF_RANGE));
	}
	
	OPControlRef& ref = this->array[h*this->v_cell_count+v];
	
	if (ref.IsValid ()) {
		ref->SetParent (OPViewNull);
		this->NotifyDetachCell (ref, h, v);
	}
	
	if (cell.IsValid ()) {
		cell->SetParent (OPCellGridRef (this));
		cell->SetFlagOkMain (FALSE);
		cell->ChangeState (OP_VIEW_IS_FROZEN, TRUE);
		this->NotifyAttachCell (cell, h, v);
	}
	
	ref = cell;
	
	this->is_minmax_ok = FALSE;
	this->is_geom_ok   = FALSE;
}



/*
 *	Return a reference to a cell specified by its horizontal
 *	and vertical array coordinates. If they are out of range,
 *	if the cell is invalid or if it is a ghost cell, simply
 *	return FALSE.
 */

Bool
OPCellGrid::GetCell (Count h, Count v, OPControlOut cell) const
{
	OP_METHOD (GetCell, 3);
	OP_ARG (1, Count, h);
	OP_ARG (2, Count, v);
	OP_ARG (3, OPControlOut, cell);
	
	if ( (h < this->h_cell_count)
	  && (v < this->v_cell_count) ) {
		cell = this->array[h*this->v_cell_count+v];
		return (cell.IsValid ()) && (cell != this->ghost_cell);
	}
	
	return FALSE;
}



/*
 *	Find a cell... Just skim through the array and look
 *	for the specified cell... Return the position in the
 *	array and TRUE if found.
 *
 *	Cerain frequent cases have been optimized.
 */

Bool
OPCellGrid::FindCell (OPControlIn cell, CountOut h, CountOut v) const
{
	OP_METHOD (FindCell, 3);
	OP_ARG (1, OPControlIn, cell);
	OP_ARG (2, CountOut, h);
	OP_ARG (3, CountOut, v);
	
	if (this->draw_cache) {
		
		//	That is what we call an educated guess : since we are currently
		//	drawing a cell, it might very well ask about its position, might
		//	it not ?
		
		if (*this->draw_cache == cell) {
			
			//	It did... Yahooo ! This did speed up the operation quite a
			//	bit and allows the cells to be stupid about their position
			//	without wasting too much time.
			
			h = this->draw_h_cache;
			v = this->draw_v_cache;
			
			return TRUE;
		}
	}


	h = 0;
	v = 0;
	
	if (array->TotalCount () == 0) {
		return FALSE;
	}
	
	Count index = 0;
	Count num   = this->h_cell_count * this->v_cell_count;
	
	while (num) {
		
		if (this->array[index] == cell) {
			return TRUE;
		}
		
		//	The cells are disposed in vertical columns. Search from first
		//	to last column...
		
		v++;
		
		if (v == this->v_cell_count) {
			h++; v = 0;
		}
		
		index++;
		num--;
	}
	
	return FALSE;
}



/*
 *	Called when the geometry changes.
 */

void
OPCellGrid::ChangeGeometry (OPRectIn box)
{
	OP_METHOD (ChangeGeometry, 1);
	OP_ARG (1, OPRectIn, box);
	
	super::ChangeGeometry (box);
	this->is_geom_ok = FALSE;
}



/*
 *	Return the cell grid's minimal and maximal size. These
 *	are computed according to the cell's min/max info.
 */

void
OPCellGrid::GetMinSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	dx = (OPCoord)(this->min_cell_dx * this->h_cell_count);
	dy = (OPCoord)(this->min_cell_dy * this->v_cell_count);
	
	if (dx < 4) dx = 4;
	if (dy < 4) dy = 4;
}

void
OPCellGrid::GetMaxSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMaxSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	dx = (OPCoord)(this->max_cell_dx * this->h_cell_count);
	dy = (OPCoord)(this->max_cell_dy * this->v_cell_count);
	
	if ( (dx > 20000) || (dx < 1) ) dx = 20000;
	if ( (dy > 20000) || (dy < 1) ) dy = 20000;
}



/*
 *	Find out which cell is under the specified point. If the
 *	cell contains a subview which matches the co-ordinates,
 *	return it.
 */

Bool
OPCellGrid::Detect (OPCoord x, OPCoord y, OPViewOut view)
{
	OP_METHOD (Detect, 3);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPViewOut, view);
	
	if (this->box.IsInside (x, y)) {
		
		//	Detect the precise cell which is at the specified co-ordinates.
		
		if (this->is_geom_ok == FALSE) {
			this->UpdateCellGeometry ();
		}
		
		OPRelCoord xx = x;
		OPRelCoord yy = y;
		
		xx -= this->box.X1 () + this->offset_x;
		yy -= this->box.Y1 () + this->offset_y;
		
		//	If out of cell array, just return ourself : the cell grid !
		
		if ((xx < 0) || (yy < 0)) {
			view = this;
			return TRUE;
		}
		
		Count h = 0;
		Count v = 0;
		
		if (this->is_isometric) {
			
			//	If the cells all have the same size, then detection is as easy
			//	as a division.
			
			h = (Count)((OPCoord)(xx) / this->cell_dx);
			v = (Count)((OPCoord)(yy) / this->cell_dy);
			
		} else {
			
			//	Find out where the mouse is...
			
			for (h = 0; h < this->h_cell_count; h++) {
				if (this->cell_ox_array[h+1] > xx) {
					break;
				}
			}
			
			for (v = 0; v < this->v_cell_count; v++) {
				if (this->cell_oy_array[v+1] > yy) {
					break;
				}
			}
		}
		
		OPControlRef cell;
		this->GetCell (h, v, cell);
		
		if ( (cell.IsValid ())
		  && (cell->Detect (x, y, view)) ) {
			
			//	The view which has been detected can be a subview of the cell.
			//	Return really the view, not the cell...
			
		} else {
			view = this;
		}
		
		return TRUE;
	}
	
	return FALSE;
}



/*
 *	Draw the cell grid. This is an optimized version of
 *	method OPView::Draw (...) ! It does not try to paint
 *	cells which do not lie in the specified rectangle.
 */

void
OPCellGrid::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	if ( (port.IsInvalid ())
	  || (rect.X1 () > this->box.X2 ())
	  || (rect.X2 () < this->box.X1 ())
	  || (rect.Y1 () > this->box.Y2 ())
	  || (rect.Y2 () < this->box.Y1 ())
	  || (this->array.IsInvalid ())
	  || (this->array->TotalCount () == 0) ) {
		return;
	}
	
	if (this->is_geom_ok == FALSE) {
		this->UpdateCellGeometry ();
	}
	
	if (what & OP_DRAW_BACKGROUND) {
		super::Draw (port, rect, OP_DRAW_BACKGROUND);
	}
	
	if (what & OP_DRAW_OBJECT) {
		
		//	Draw the foreground...
		
		if (this->draw_frame_high) {
			port->ReturnAdorner ()->DrawRectHigh (this->box);
		}
		
		if (this->draw_frame_low) {
			port->ReturnAdorner ()->DrawRectLow (this->box);
		}
	}
	
	OPRect old_rect;
	port->GetClipRect (old_rect);

	if (this->is_isometric == FALSE) {
		
		//	Each cell might have a different position and size. Look for each cell
		//	in turn and decide whether it needs to be drawn or not.
		
		OPControlRef cell;
		OPRect       box;
		
		for (Count x = 0; x < this->h_cell_count; x++) {
			for (Count y = 0; y < this->v_cell_count; y++) {
				
				this->GetCellRect (x, y, box);
				
				if (box.DoOverlap (rect)) {
					if (this->GetCell (x, y, cell)) {
						
						//	The cell which we are going to draw might in turn ask about
						//	its position. We have better take a note about it to speed
						//	things up in case "FindCell" is called.
						
						this->draw_h_cache = x;
						this->draw_v_cache = y;
						
						box *= rect;
						port->SetClipRect (box);
						cell->Draw (port, box, what);
					}
				}
			}
		}
		
	} else {	
		
		//	Each cell has the same size, therefore we need not examine each
		//	cell to know if it will have to be drawn or not.
		
		OPRect new_rect;
		
		OPCoord offset_before_x = rect.X1 () - this->box.X1 () - this->offset_x;
		OPCoord offset_before_y = rect.Y1 () - this->box.Y1 () - this->offset_y;
		OPCoord offset_after_x  = rect.X2 () - this->box.X1 () - this->offset_x + this->cell_dx - 1;
		OPCoord offset_after_y  = rect.Y2 () - this->box.Y1 () - this->offset_y + this->cell_dy - 1;
		
		if (offset_before_x < 0) offset_before_x = 0;
		if (offset_before_y < 0) offset_before_y = 0;
		if (offset_after_x < 0) offset_after_x = 0;
		if (offset_after_y < 0) offset_after_y = 0;
		
		Count ix = (Count)(offset_before_x / this->cell_dx);
		Count iy = (Count)(offset_before_y / this->cell_dy);
		Count nx = (Count)(offset_after_x / this->cell_dx);
		Count ny = (Count)(offset_after_y / this->cell_dy);
		
		Count x;
		Count y;
		OPControlRef cell;
		
		//	The cell reference will not change while we draw, but the cell pointed
		//	to by it will vary. Just take the address once.
		
		this->draw_cache = & cell;
		
		//	We know exactly which cells need to be drawn, so we do not use the more
		//	general drawing algorithm used by the View class, which might produce
		//	lots of empty clipping rectangles.
		
		for (x = ix; x <= nx; x++) {
			for (y = iy; y <= ny; y++) {
				
				if ( (this->GetCell (x, y, cell))
				  && (cell.IsValid ()) ) {
					
					//	The cell which we are going to draw might in turn ask about
					//	its position. We have better take a note about it to speed
					//	things up in case "FindCell" is called.
					
					this->draw_h_cache = x;
					this->draw_v_cache = y;
					
					new_rect  = rect;
					new_rect *= cell->ReturnBox ();
					
					port->SetClipRect (new_rect);
					cell->Draw (port, new_rect, what);
				}
			}
		}
	}
		
	this->draw_cache   = 0;
	this->draw_h_cache = 0;
	this->draw_v_cache = 0;

	//	Finally, restore original clipping rectangle. This might even
	//	be suppressed !
	
	port->SetClipRect (old_rect);
}

