/*
 *	colorman.cxx
 *
 *	OPaC Grail color manager : the internal OPaC color model is
 *	based on a 32-bit true color alpha + RGB model.
 *
 *	White is set to 1,1,1 and black to 0,0,0. An opaque color has an
 *	alpha value of 0; a transparent color has an alpha value of 1.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/grail.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPColorMan);

/****************************************************************/

OPColorManRef	OPColorMan::current = (OPColorMan*)(0);

/****************************************************************/


/*
 *	Initialise a color manager.
 */	

OPColorMan::OPColorMan ()
{
}

/* 
 *	Clean up the color manager.
 */

OPColorMan::~OPColorMan ()
{
}

/*
 *	Save the color manager's state.
 */

void
OPColorMan::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	//	This is silly... Why save the color manager ?
	
	storage << OPStorageSync::mark;
}



/*
 *	Restore the color manager's state.
 */

void
OPColorMan::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> OPStorageSync::mark;
}



/*
 *	Find the nearest matching pixel color for the specified
 *	RGB color. The components are in the 0..1 range.
 */

OPColorPixel
OPColorMan::FindColor (Real r, Real g, Real b, Real alpha) const
{
	OP_METHOD (FindColor, 4);
	OP_ARG (1, Real, r);
	OP_ARG (2, Real, g);
	OP_ARG (3, Real, b);
	OP_ARG (4, Real, alpha);
	
	Card16 rval = (Card16)(r * 256);
	Card16 gval = (Card16)(g * 256);
	Card16 bval = (Card16)(b * 256);
	Card16 aval = (Card16)(alpha * 256);
	if (rval > 255) rval = 255;
	if (gval > 255) gval = 255;
	if (bval > 255) bval = 255;
	if (aval > 255) aval = 255;
	
	return (OPColorPixel)((aval << 24) | (rval << 16) | (gval << 8) | (bval));
}



/*
 *	Find the nearest matching pixel color for the specified
 *	RGB color. The components are in the 0..255 range.
 */

OPColorPixel
OPColorMan::FindColor (Card8 r, Card8 g, Card8 b, Card8 alpha) const
{
	OP_METHOD (FindColor, 4);
	OP_ARG (1, Card8, r);
	OP_ARG (2, Card8, g);
	OP_ARG (3, Card8, b);
	OP_ARG (4, Card8, alpha);
	
	return (OPColorPixel)((alpha << 24) | (r << 16) | (g << 8) | (b));
}



/*
 *	Find the nearest matching pixel color for the specified
 *	gray level. The level is in the range 0..1.
 */

OPColorPixel
OPColorMan::FindColor (Real i, Real alpha) const
{
	OP_METHOD (FindColor, 2);
	OP_ARG (1, Real, i);
	OP_ARG (2, Real, alpha);
	
	return this->FindColor (i, i, i, alpha);
}



/*
 *	Find the nearest matching pixel color for the specified
 *	gray level. The level is in the range 0..255.
 */

OPColorPixel
OPColorMan::FindColor (Card8 i, Card8 alpha) const
{
	OP_METHOD (FindColor, 2);
	OP_ARG (1, Card8, i);
	OP_ARG (2, Card8, alpha);
	
	return this->FindColor (i, i, i, alpha);
}

/*
 *	Return the color components from a pixel value.
 */

Bool
OPColorMan::FindColorRGB (OPColorPixel p, RealOut r, RealOut g, RealOut b) const
{
	OP_METHOD (FindColorRGB, 4);
	OP_ARG (1, OPColorPixel, p);
	OP_ARG (2, RealOut, r);
	OP_ARG (3, RealOut, g);
	OP_ARG (4, RealOut, b);
	
	Real alpha = 0.0;
	return this->FindColorRGB (p, r, g, b, alpha);
}

Bool
OPColorMan::FindColorRGB (OPColorPixel p, RealOut r, RealOut g, RealOut b, RealOut alpha) const
{
	OP_METHOD (FindColorRGB, 5);
	OP_ARG (1, OPColorPixel, p);
	OP_ARG (2, RealOut, r);
	OP_ARG (3, RealOut, g);
	OP_ARG (4, RealOut, b);
	OP_ARG (5, RealOut, alpha);
	
	Card8 r8 = 0;
	Card8 g8 = 0;
	Card8 b8 = 0;
	Card8 a8 = 0;
	
	if (this->FindColorRGB (p, r8, g8, b8, a8)) {
		
		r = r8 / 255.0;
		g = g8 / 255.0;
		b = b8 / 255.0;
		alpha = a8 / 255.0;
		
		return TRUE;
	}
	
	return FALSE;
}


/*
 *	Return the color components from a pixel value.
 */

Bool
OPColorMan::FindColorRGB (OPColorPixel p, Card8Out r, Card8Out g, Card8Out b) const
{
	OP_METHOD (FindColorRGB, 4);
	OP_ARG (1, OPColorPixel, p);
	OP_ARG (2, Card8Out, r);
	OP_ARG (3, Card8Out, g);
	OP_ARG (4, Card8Out, b);
	
	Card8 alpha = 0;
	
	return this->FindColorRGB (p, r, g, b, alpha);
}
	
Bool
OPColorMan::FindColorRGB (OPColorPixel p, Card8Out r, Card8Out g, Card8Out b, Card8Out a) const
{
	OP_METHOD (FindColorRGB, 5);
	OP_ARG (1, OPColorPixel, p);
	OP_ARG (2, Card8Out, r);
	OP_ARG (3, Card8Out, g);
	OP_ARG (4, Card8Out, b);
	OP_ARG (5, Card8Out, a);
	
	Card32 pix32 = (Card32)(p);
	
	r = (Card8)((pix32 & 0x000000FF) >> 0);
	g = (Card8)((pix32 & 0x0000FF00) >> 8);
	b = (Card8)((pix32 & 0x00FF0000) >> 16);
	a = (Card8)((pix32 & 0xFF000000) >> 24);
	
	return TRUE;
}

/****************************************************************/

/*
 *	Just change the alpha channel of the pixel.
 */

OPColorPixel
OPColorMan::SetAlpha (OPColorPixel pixel, Card8 alpha) const
{
	OP_METHOD (SetAlpha, 2);
	OP_ARG (1, OPColorPixel, pixel);
	OP_ARG (2, Card8, alpha);
	
	return (alpha << 24) | (pixel & 0x00FFFFFF);
}

OPColorPixel
OPColorMan::SetAlpha (OPColorPixel pixel, Real alpha) const
{
	OP_METHOD (SetAlpha, 2);
	OP_ARG (1, OPColorPixel, pixel);
	OP_ARG (2, Real, alpha);
	
	Card16 a = (Card16)(alpha * 256);
	if (a > 255) a = 255;
	
	return (a << 24) | (pixel & 0x00FFFFFF);
}

/****************************************************************/

/*
 *	Initialise the default system color manager. This will
 *	be accepted only once.
 */

void
OPColorMan::InitSystemColorMan (OPColorManIn manager)
{
	OP_MEMBER (InitSystemColorMan, 1);
	OP_ARG (1, OPColorManIn, manager);
	
	OP_ASSERT (OPColorMan::current.IsInvalid ());
	
	OPColorMan::current = manager;
}

void
OPColorMan::KillSystemColorMan ()
{
	OP_MEMBER (KillSystemColorMan, 0);
	
	OPColorMan::current = 0;
}

/****************************************************************/
