/*
 *	iconset.cxx
 *
 *	The Icon Set class draws several icons side by side.
 *
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/iconset.h>
#include <OPaC/new.h>
#include <OPaC/lookman.h>
#include <OPaC/grail.h>
#include <OPaC/array.h>

/****************************************************************/

OP_IMPLEMENTATION (iconset.h);
OP_IMPLEMENT_CLASS (OPIconSet);

/****************************************************************/


/*
 *	Create an (empty) icon set.
 */

OPIconSet::OPIconSet ()
{
	OP_METHOD (OPIconSet, 0);

	this->icons = OP_NEW (OPArray);
}



/*
 *	Destroy an icon set.
 */

OPIconSet::~OPIconSet ()
{
	OP_METHOD (~OPIconSet, 0);
}



/*
 *	Save an icon set to the storage.
 */

void
OPIconSet::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	super::StoreInstance (storage);
	
	storage << this->icons;
	storage << OPStorageSync::mark;
}



/*
 *	Restore an icon set from the storage.
 */

void
OPIconSet::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	super::RestoreInstance (storage);
	
	storage >> this->icons;
	storage >> OPStorageSync::mark;
}



/*
 *	Prepare the icon set for its destruction.
 */

void
OPIconSet::Delete ()
{
	OP_METHOD (Delete, 0);
	
	Count max = this->icons.IsValid () ? this->icons->TotalCount () : 0;
	
	while (max--) {
		if (this->icons[max].IsValid ()) {
			this->icons[max]->SetParent (OPViewNull);
			this->icons[max] = OPIconNull;
		}
	}

	this->icons->SetSize (0);
	
	super::Delete ();
}



/*
 *	Define an icon in the set.
 */

Bool
OPIconSet::InsertIcon (Count index, OPLookIcon icon)
{
	OP_METHOD (InsertIcon, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPLookIcon, icon);
	
	return this->InsertIcon (index, icon, icon, icon, icon);
}

Bool
OPIconSet::InsertIcon (Count index, OPLookIcon icon, OPLookIcon press, OPLookIcon hilite, OPLookIcon dim)
{
	OP_METHOD (InsertIcon, 5);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPLookIcon, icon);
	OP_ARG (3, OPLookIcon, press);
	OP_ARG (4, OPLookIcon, hilite);
	OP_ARG (5, OPLookIcon, dim);
	
	OPIconRef i = OP_NEW (OPIcon);
	i->Initialise (icon, press, hilite, dim);
	i->SetDefaultSize ();
	return this->InsertIcon (index, i);
}

Bool
OPIconSet::InsertIcon (Count index, OPIconIn icon)
{
	OP_METHOD (InsertIcon, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPIconIn, icon);
	
	this->icons->Insert (index);
	this->icons[index] = icon;
	
	icon->SetParent (this);

	this->UpdateGeometry ();
	
	return TRUE;
}


/*
 *	Update the positions of the icons and the size of the
 *	whole icon set.
 */

void
OPIconSet::UpdateGeometry ()
{
	OP_METHOD (UpdateGeometry, 0);
	
	OPCoord dx = 0;
	OPCoord dy = 0;
	
	this->GetMinSize (dx, dy);
	this->SetSize (dx, dy);
	
	OPCoord x = this->box.X1 ();
	OPCoord y = this->box.Y1 ();
	
	Count pos = 0;
	Count max = this->icons->TotalCount ();
	
	while (pos < max) {
		if (this->icons[pos].IsValid ()) {
			this->icons[pos]->SetOrigin (x, y);
			this->icons[pos]->GetSize (dx, dy);
			x += dx;
		}
		pos++;
	}
}


/*
 *	Remove an icon from the set.
 */

Bool
OPIconSet::RemoveIcon (Count index)
{
	OP_METHOD (RemoveIcon, 1);
	OP_ARG (1, Count, index);
	
	if (this->icons[index].IsValid ()) {
		this->icons[index]->SetParent (OPViewNull);
		this->icons[index] = 0;
		this->UpdateGeometry ();
	}
	
	return (this->icons->Remove (index));
}


/*
 *	Return the icon's minimal size. This will be the
 *	glyph's size.
 */

void
OPIconSet::GetMinSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	OPCoord tdx = 0;
	OPCoord tdy = 0;
	
	Count pos = 0;
	Count max = this->icons->TotalCount ();
	
	dx = 0;
	dy = 0;
	
	while (pos < max) {
		if (this->icons[pos].IsValid ()) {
			this->icons[pos]->GetMinSize (tdx, tdy);
			dx += tdx;
			if (dy < tdy) dy = tdy;
		}
		pos++;
	}
}


/*
 *	Return the icon's maximum size. This is the same
 *	as the minimum size !
 */

void
OPIconSet::GetMaxSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMaxSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	OPCoord tdx = 0;
	OPCoord tdy = 0;
	
	Count pos = 0;
	Count max = this->icons->TotalCount ();
	
	dx = 0;
	dy = 0;
	
	while (pos < max) {
		if (this->icons[pos].IsValid ()) {
			this->icons[pos]->GetMaxSize (tdx, tdy);
			dx += tdx;
			if (dy < tdy) dy = tdy;
		}
		pos++;
	}
}
