/*
 *	ieeemath.cxx
 *	
 *	The IEEE double cannot be saved as is, because its size is
 *	very architecture dependent. That's why some portable format
 *	has been defined :
 *
 *	- 31-bit exponent + mantissa sign bit
 *	- 64-bit mantissa
 *	
 *	CAUTION :
 *
 *	This implementation currently does not support infinity, NaN
 *	and denormalized IEEE values !
 *	
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/ieeemath.h>
#include <OPaC/object.h>

/****************************************************************/

union IEEE_Decode
{
	double	real;
	Card8	data[sizeof (double)];
};

/****************************************************************/

static Count ieee_exp_size    = 0;
static Count ieee_exp_shift   = 0;
static Count ieee_man_size    = 0;

static Count ieee_man_shift_h = 0;
static Count ieee_man_size_h  = 0;
static Count ieee_man_shift_l = 0;
static Count ieee_man_size_l  = 0;

static Count ieee_sign        = 0;

/****************************************************************/

/*
 *	Extract some bits from a huge integer value, specified
 *	as an array of "len" bytes.
 *	
 *	For example, extracting data at pos=3 and width=6 of
 *	the hexadecimal value 12 34 56 78 (big endian) or value
 *	78 56 34 12 (little endian) will yield 001111 !
 *
 *	The data in the array must be stored in the machine's
 *	native endian format.
 */

static Card32
OPaC_ExtractBits (const Card8* data, Count len, Count pos, Count width)
{
	OP_FUNCTION (OPaC_ExtractBits, 4);
	OP_ARG (1, const Card8*, data);
	OP_ARG (2, Count, len);
	OP_ARG (3, Count, pos);
	OP_ARG (4, Count, width);
	
	Card32 result = 0;
	Card32 index  = pos / 8;
	Card32 shift  = pos - index * 8;
	Card32 move   = 0;
	
#ifdef SYS_BIG_ENDIAN
	
	while (width--) {
		
		result |= ((Card32)((data[len - index - 1] >> shift) & 1)) << move;
		move   += 1;
		shift  += 1;
		
		if (shift == 8) {
			index++;
			shift = 0;
		}
	}
	
#else  /* SYS_LITTLE_ENDIAN */
	
	while (width--) {
		
		result |= ((Card32)((data[index] >> shift) & 1)) << move;
		move   += 1;
		shift  += 1;
		
		if (shift == 8) {
			index++;
			shift = 0;
		}
	}
	
#endif /* SYS_LITTLE_ENDIAN */
	
	return result;
}


/*
 *	This is the opposite of the ExtractBits function and
 *	will insert the "result" into the "data" array of
 *	bits.
 */

static void
OPaC_InsertBits (Card32 result, Card8* data, Count len, Count pos, Count width)
{
	OP_FUNCTION (OPaC_InsertBits, 5);
	OP_ARG (1, Card32, result);
	OP_ARG (2, Card8*, data);
	OP_ARG (3, Count, len);
	OP_ARG (4, Count, pos);
	OP_ARG (5, Count, width);
	
	Card32 index  = pos / 8;
	Card32 shift  = pos - index * 8;
	Card32 move   = 0;
	
#ifdef SYS_BIG_ENDIAN
	
	while (width--) {
		
		data[len - index - 1] |= ((Card32)((result >> move) & 1)) << shift;
		move  += 1;
		shift += 1;
		
		if (shift == 8) {
			index++;
			shift = 0;
		}
	}
	
#else  /* SYS_LITTLE_ENDIAN */
	
	while (width--) {
		
		data[index] |= ((Card32)((result >> move) & 1)) << shift;
		move        += 1;
		shift       += 1;
		
		if (shift == 8) {
			index++;
			shift = 0;
		}
	}
	
#endif /* SYS_LITTLE_ENDIAN */
}


/*
 *	Find out where this machine's IEEE exponent, mantissa
 *	and sign bits reside.
 */

static void
OPaC_DemistifyDouble ()
{
	OP_FUNCTION (OPaC_DemistifyDouble, 0);
	
	IEEE_Decode one;
	IEEE_Decode one_five;
	IEEE_Decode minus_one;
	
	Count       index;
	
	one.real       =  1.0;
	one_five.real  =  1.5;
	minus_one.real = -1.0;
	
	//	Clear the exponent bits.
	
	for (index = 0; index < sizeof (double); index++) {
		one_five.data[index]  = one_five.data[index]  ^ one.data[index];
		minus_one.data[index] = minus_one.data[index] ^ one.data[index];
	}
	
	//	Find the sign bit and remember its position.
	
	for (index = 0; index < sizeof (double) * 8; index++)
		if (OPaC_ExtractBits (minus_one.data, sizeof (double), index, 1))
			break;
	
	ieee_sign = index;
	
	//	Find the exponent position and implicitely, its size.
	
	for (index = 0; index < sizeof (double) * 8; index++)
		if (OPaC_ExtractBits (one.data, sizeof (double), index, 1))
			break;
	
	ieee_exp_shift = index;
	ieee_exp_size  = ieee_sign - ieee_exp_shift;
	
	//	The mantissa sign bit should be merged with the exponent,
	//	according to IEEE. Check that to be sure !
	
	OP_ASSERT ((ieee_exp_size + ieee_exp_shift) == ieee_sign);
	
	//	Find the width of the mantissa.
	
	for (index = 0; index < sizeof (double) * 8; index++)
		if (OPaC_ExtractBits (one_five.data, sizeof (double), index, 1))
			break;
	
	ieee_man_size = index + 1;
	
	ieee_man_shift_h = ieee_man_size - 32;
	ieee_man_size_h  = 32;
	if (ieee_man_size < 64) ieee_man_shift_l = 0; else ieee_man_shift_l = ieee_man_size - 64;
	if (ieee_man_size < 64) ieee_man_size_l  = ieee_man_size - 32; else ieee_man_size_l = 32;
}



/*
 *	Export the specified double value in a portable, non
 *	IEEE, format. This stores the 31-bit exponent, the
 *	mantissa sign bit and the 64-bit mantissa.
 */

void
OPaC_ExportDouble (double value, Card32& exponent, Card32* mantissa)
{
	OP_FUNCTION (OPaC_ExportDouble, 3);
	OP_ARG (1, double, value);
	OP_ARG (2, Card32&, exponent);
	OP_ARG (3, Card32*, mantissa);
	
	if (ieee_exp_size == 0) OPaC_DemistifyDouble ();
	
	Card8* data = (Card8*)(&value);
	Int32  exp  = (Int32)  OPaC_ExtractBits (data, sizeof (double), ieee_exp_shift, ieee_exp_size);
	Card32 sign = (Card32) OPaC_ExtractBits (data, sizeof (double), ieee_sign, 1);
	
	exp = exp - (1 << (ieee_exp_size - 1));
	
	exponent    = (Card32)(exp & 0x7FFFFFFF) | (sign << 31);
	mantissa[0] = (Card32) OPaC_ExtractBits (data, sizeof (double), ieee_man_shift_h, ieee_man_size_h);
	mantissa[1] = ((Card32) OPaC_ExtractBits (data, sizeof (double), ieee_man_shift_l, ieee_man_size_l)) << (32 - ieee_man_size_l);
}



/*
 *	Import the specified double value from a portable
 *	format (31-bit exponent, mantissa sign bit and a
 *	64-bit mantissa).
 */

void
OPaC_ImportDouble (double& value, Card32 exponent, const Card32* mantissa)
{
	OP_FUNCTION (OPaC_ImportDouble, 3);
	OP_ARG (1, double&, value);
	OP_ARG (2, Card32, exponent);
	OP_ARG (3, const Card32*, mantissa);
	
	if (ieee_exp_size == 0) OPaC_DemistifyDouble ();
	
	Card8* data = (Card8*)(&value);
	for (int i = 0; i < sizeof (double); i++) data[i] = 0;
	
	Int32  exp  = ((((Int32)(exponent & 0x7FFFFFFF)) << 1) >> 1) + (1 << (ieee_exp_size - 1));
	Card32 sign = (exponent >> 31);
	
	OPaC_InsertBits (exp,         data, sizeof (double), ieee_exp_shift, ieee_exp_size);
	OPaC_InsertBits (sign,        data, sizeof (double), ieee_sign, 1);
	OPaC_InsertBits (mantissa[0], data, sizeof (double), ieee_man_shift_h, ieee_man_size_h);
	OPaC_InsertBits (mantissa[1] >> (32 - ieee_man_size_l), data, sizeof (double), ieee_man_shift_l, ieee_man_size_l);
}

