/*
 *	image.cxx
 *
 *	The Image class displays an image (useful for backgrounds
 *	and large icons).
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/grail.h>
#include <OPaC/image.h>
#include <OPaC/new.h>
#include <OPaC/string.h>

#include <OPaC/pubmacro.h>
#include <OPaC/targetmacro.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPImage);

/****************************************************************/

OP_PUBLISH_MODIFUNC (OPImage, "IsCentered", IsImageCentered, SetCenteredImage, "Centered image", Bool, Bool);
OP_PUBLISH_MODIFUNC (OPImage, "IsTiled", IsImageTiled, SetTiledImage, "Tiled image", Bool, Bool);
OP_PUBLISH_MODIFIER (OPImage, "ImageName", GetImageName, SetImageName, "Image name", OPObject, OPString);

/****************************************************************/

OP_TARGET_ACTOR (OPImage, SetDefaultSize);

OP_TARGET_CLASS (OPImage);
OP_TARGET_ARG (OPStringRef, name, "Image name");
OP_TARGET_ARG (Bool, is_centered, "Centered image");
OP_TARGET_ARG (Bool, is_tiled, "Tiled image");
OP_TARGET_CALL (Initialise (name, is_centered, is_tiled));
OP_TARGET_END ();

/****************************************************************/

OPImage::OPImage ()
{
	OP_METHOD (OPImage, 0);
	
	this->pixmap = OPGrailPixmapNull;
	this->name   = OPStringNull;
	this->is_centered = TRUE;
	this->is_tiled    = FALSE;
}

OPImage::~OPImage ()
{
	OP_METHOD (~OPImage, 0);
}

/****************************************************************/

/*
 *	Save an image to the storage.
 */

void
OPImage::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->name << this->is_centered << this->is_tiled;
	storage << OPStorageSync::mark;
}



/*
 *	Restore an image from the storage.
 */

void
OPImage::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->name >> this->is_centered >> this->is_tiled;
	storage >> OPStorageSync::mark;
}



/*
 *	Prepare the image for its destruction.
 */

void
OPImage::Delete ()
{
	OP_METHOD (Delete, 0);
	
	super::Delete ();
	
	this->name = OPStringNull;
	
	if (this->pixmap.IsValid ()) {
		this->pixmap->Delete ();
		this->pixmap = 0;
	}
}


/*
 *	Awake the object after it has been restored.
 */

void
OPImage::Awake ()
{
	OP_METHOD (Awake, 0);
	
	super::Awake ();
		
	this->pixmap = OPaC_LoadGIF (this->name, this->ReturnMemZone ());
}


Bool
OPImage::Initialise (OPStringIn name, Bool center, Bool tile)
{
	OP_METHOD (Initialise, 3);
	OP_ARG (1, OPStringIn, name);
	OP_ARG (2, Bool, center);
	OP_ARG (3, Bool, tile);
	
	if (name.IsInvalid ()) {
		return FALSE;
	}
	
	this->pixmap = OPaC_LoadGIF (name, this->ReturnMemZone ());
	
	if (this->pixmap.IsValid ()) {
		
		this->name        = name->ReturnString ();
		this->is_centered = center;
		this->is_tiled    = tile;
		
		return TRUE;
	}
	
	return FALSE;
}


Bool
OPImage::SetDefaultSize ()
{
	OP_METHOD (SetDefaultSize, 0);
	
	if (this->pixmap.IsValid ()) {
		
		OPCoord dx;
		OPCoord dy;
		
		this->GetImageSize (dx, dy);
		this->SetSize (dx, dy);
		
		return TRUE;
	}
	
	return FALSE;
}


Bool
OPImage::SetImageName (OPStringIn name)
{
	OP_METHOD (SetImageName, 1);
	OP_ARG (1, OPStringIn, name);
	
	if (this->pixmap.IsValid ()) {
		this->pixmap->Delete ();
		this->pixmap = 0;
	}
	
	return this->Initialise (name, this->is_centered, this->is_tiled);
}

Bool
OPImage::SetCenteredImage (Bool yes)
{
	OP_METHOD (SetCenteredImage, 1);
	OP_ARG (1, Bool, yes);
	
	this->is_centered = yes;
	
	return TRUE;
}

Bool
OPImage::SetTiledImage (Bool yes)
{
	OP_METHOD (SetTiledImage, 1);
	OP_ARG (1, Bool, yes);
	
	this->is_tiled = yes;
	
	return TRUE;
}

Bool
OPImage::GetImageName (OPStringOut name) const
{
	OP_METHOD (GetImageName, 1);
	OP_ARG (1, OPStringOut, name);
	
	name = this->name;
	
	return TRUE;
}

Bool
OPImage::GetImageSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetImageSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	if (this->pixmap.IsValid ()) {
		this->pixmap->GetSize (dx, dy);
		return TRUE;
	}
	
	return FALSE;
}


Bool
OPImage::IsImageCentered () const
{
	OP_METHOD (IsImageCentered, 0);
	
	return this->is_centered;
}

Bool
OPImage::IsImageTiled () const
{
	OP_METHOD (IsImageTiled, 0);
	
	return this->is_tiled;
}

void
OPImage::Draw (OPGrailPortIn port, OPRectIn box, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, box);
	OP_ARG (3, OPDrawOp, what);
	
	if ( (what & OP_DRAW_BACKGROUND)
	  && (this->pixmap.IsValid ()) ) {
		
		OPCoord dx;
		OPCoord dy;
		this->GetImageSize (dx, dy);
		
		OP_ASSERT (dx > 0);
		OP_ASSERT (dy > 0);
		
		OPCoord idx = dx;
		OPCoord idy = dy;
		OPCoord ox  = this->box.X1 ();
		OPCoord oy  = this->box.Y1 ();
		Count   nx  = 1;
		Count   ny  = 1;
		
		//	If the image has to be tiled, find out how many stripes we will have
		//	to display in order to fill the view's surface.
		
		if (this->is_tiled) {
			while (idx < this->box.DX ()) {
				idx += dx;
				nx++;
			}
			while (idy < this->box.DY ()) {
				idy += dy;
				ny++;
			}
		}
		
		if (this->is_centered) {
			
			if (this->is_tiled) {
				if ((nx & 1) == 0) {
					idx += dx;
					nx++;
				}
				if ((ny & 1) == 0) {
					idy += dy;
					ny++;
				}
			}
			
			ox += (this->box.DX () - idx) / 2;
			oy += (this->box.DY () - idy) / 2;
		}
		
		OPRect box (0, 0, dx-1, dy-1);
		
		for (Count y = 0; y < ny; y++) {
			for (Count x = 0; x < nx; x++) {
				box.SetOrigin (ox+x*dx, oy+y*dy);
				port->CopyRect (this->pixmap, 0, 0, box);
			}
		}
	}
	
	super::Draw (port, box, what);
}
