/*
 *	multiview.cxx
 *
 *	The Multiple View is a base class for building tabbed views
 *	and step-by-step wizard views.
 *	visible at a given time.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/container.h>
#include <OPaC/multiview.h>
#include <OPaC/new.h>

#include <OPaC/targetmacro.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPMultiView);

/****************************************************************/

OP_TARGET_ACTOR (OPMultiView, ActivateNextView);
OP_TARGET_ACTOR (OPMultiView, ActivatePrevView);

/****************************************************************/


OPMultiView::OPMultiView ()
{
	OP_METHOD (OPMultiView, 0);
	
	this->active_index = 0;
}

OPMultiView::~OPMultiView ()
{
	OP_METHOD (~OPMultiView, 0);
}



/*
 *	Adjust the overall size so that the largest view
 *	fits exactly into the containing "multi-view".
 */

void
OPMultiView::AdjustOverallSize ()
{
	OP_METHOD (AdjustOverallSize, 0);
	
	OPCoord dx;
	OPCoord dy;
	
	this->GetMinSize (dx, dy);
	this->SetSize (dx, dy);
}

/*
 *	Return the control's minimal size (depends both on
 *	the fixed minimum and on the super-view's minimum).
 */

void
OPMultiView::GetMinSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	super::GetMinSize (dx, dy);
	
	if (this->views.IsValid ()) {
		Count     index = 0;
		OPViewRef view;
		
		while (this->views->Get (index++, view)) {
			OPCoord dx1;
			OPCoord dy1;
			if (view.IsValid ()) {
				view->GetMinSize (dx1, dy1);
				if (dx1 > dx) dx = dx1;
				if (dy1 > dy) dy = dy1;
			}
		}
	}
}


/*
 *	Return the control's maximum size. Return an arbitrary
 *	huge value.
 */

void
OPMultiView::GetMaxSize (OPCoordOut dx, OPCoordOut dy) const
{
	OP_METHOD (GetMinSize, 2);
	OP_ARG (1, OPCoordOut, dx);
	OP_ARG (2, OPCoordOut, dy);
	
	super::GetMaxSize (dx, dy);
	
	if (this->views.IsValid ()) {
		Count     index = 0;
		OPViewRef view;
		
		while (this->views->Get (index++, view)) {
			OPCoord dx1;
			OPCoord dy1;
			if (view.IsValid ()) {
				view->GetMaxSize (dx1, dy1);
				if (dx1 < dx) dx = dx1;
				if (dy1 < dy) dy = dy1;
			}
		}
	}
}


/*
 *	Save the view to the storage.
 */

void
OPMultiView::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->views;
	storage << this->active_view << this->active_index;
	
	storage << OPStorageSync::mark;
}



/*
 *	Restore the view from the storage.
 */

void
OPMultiView::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->views;
	storage >> this->active_view >> this->active_index;
	
	storage >> OPStorageSync::mark;
}


void
OPMultiView::ChangeGeometry (OPRectIn box)
{
	OP_METHOD (ChangeGeometry, 1);
	OP_ARG (1, OPRectIn, box);
	
	super::ChangeGeometry (box);
	
	this->UpdateGeometry (box);
}

void
OPMultiView::UpdateGeometry (OPRectIn box)
{
	OP_METHOD (UpdateGeometry, 1);
	OP_ARG (1, OPRectIn, box);
	
	if (this->active_view.IsValid ()) {
		this->active_view->SetBox (box);
	}
}


void
OPMultiView::Delete ()
{
	OP_METHOD (Delete, 0);
	
	OPViewRef view;
	Count     index = 0;
	
	if (this->views.IsValid ()) {
		while (this->views->Get (index++, view)) {
			if (view.IsValid ()) {
				view->SetParent (OPViewNull);
				view->Delete ();
			}
		}
	}
	
	this->views->Delete ();
	
	super::Delete ();
	
	this->views       = OPListNull;
	this->active_view = OPViewNull;
}


/*
 *	Make a specified view active. An index of 0 means that
 *	no view is currently active.
 */

Bool
OPMultiView::ActivateView (Count index)
{
	OP_METHOD (ActivateView, 1);
	OP_ARG (1, Count, index);
	
	OPViewRef view = OPViewNull;
	
	if ( (this->views.IsValid ())
	  && ((index == 0) || (this->views->Get (index-1, view)))
	  && (this->MakeViewActive (view, index)) ) {
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPMultiView::ActivateView (OPViewIn view)
{
	OP_METHOD (ActivateView, 1);
	OP_ARG (1, OPViewIn, view);
	
	Count index = 0;
	
	if ( (this->views.IsValid ())
	  && (this->views->Find (view, index))
	  && (this->MakeViewActive (view, index+1)) ) {
		return TRUE;
	}
	
	return FALSE;
}

Bool
OPMultiView::ActivateNextView ()
{
	OP_METHOD (ActivateNextView, 0);
	
	return this->ActivateView (this->active_index+1);
}

Bool
OPMultiView::ActivatePrevView ()
{
	OP_METHOD (ActivatePrevView, 0);
	
	if (this->active_index == 0) {
		return FALSE;
	}
	
	return this->ActivateView (this->active_index-1);
}



/*
 *	The active view index starts at 1. 0 means no view is
 *	active.
 */

Bool
OPMultiView::MakeViewActive (OPViewIn view, Count index)
{
	OP_METHOD (MakeViewActive, 2);
	OP_ARG (1, OPViewIn, view);
	OP_ARG (2, Count, index);
	
	if ( (this->active_view == view)
	  && (this->active_index == index) ) {
		return TRUE;
	}
	
	if (view.IsValid ()) {
		
		if (this->active_view.IsValid ()) {
			this->active_view->SetParent (OPViewNull);
		}
		
		this->active_index = index;
		this->active_view  = view;
		
		if (this->active_view.IsValid ()) {
			this->active_view->SetParent (this);
			this->active_view->SetBox (this->box);
		}
		
		return TRUE;
	}
	
	return FALSE;
}


Count
OPMultiView::ReturnActiveView () const
{
	OP_METHOD (ReturnActiveView, 0);
	return this->active_index;
}

Bool
OPMultiView::GetActiveView (OPViewOut view) const
{
	OP_METHOD (GetActiveView, 1);
	OP_ARG (1, OPViewOut, view);
	
	view = this->active_view;
	
	return view.IsValid ();
}

Bool
OPMultiView::GetNthView (Count nth, OPViewOut view) const
{
	OP_METHOD (GetNthView, 2);
	OP_ARG (1, Count, nth);
	OP_ARG (2, OPViewOut, view);
	
	if ( (this->views.IsValid ())
	  && (nth > 0)
	  && (this->views->Get (nth-1, view)) ) {
		return TRUE;
	}
	
	return FALSE;
}


Bool
OPMultiView::SetViewList (OPListIn views)
{
	OP_METHOD (SetViewList, 1);
	OP_ARG (1, OPListIn, views);
	
	this->views = views;
	this->active_view  = OPViewNull;
	this->active_index = 0;
	
	return TRUE;
}

Bool
OPMultiView::GetViewList (OPListOut views) const
{
	OP_METHOD (GetViewList, 1);
	OP_ARG (1, OPListOut, views);
	
	views = this->views;
	
	return TRUE;
}

Count
OPMultiView::CountViews () const
{
	OP_METHOD (CountViews, 0);
	return this->views.IsValid () ? this->views->TotalCount () : 0;
}
