/*
 *	objectpane.cxx
 *
 *	The Object Pane class represents objects in a grid, in a visual
 *	way (like files in a browser).
 *	
 *	(C) Copyright 1997-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/app.h>
#include <OPaC/boxview.h>
#include <OPaC/cellgrid.h>
#include <OPaC/grail.h>
#include <OPaC/icon.h>
#include <OPaC/lookman.h>
#include <OPaC/new.h>
#include <OPaC/objectbrowser.h>
#include <OPaC/objectpane.h>
#include <OPaC/scrview.h>
#include <OPaC/selgrid.h>
#include <OPaC/string.h>
#include <OPaC/textstyle.h>
#include <OPaC/texttag.h>

#include <OPaC/pubmacro.h>
#include <OPaC/targetmacro.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPObjectPane);

OP_PUBLISH_MODIFIER (OPObjectPane, "Provider", GetProvider, SetProvider, "Provider", OPObject, OPGenericProvider);

/****************************************************************/


Bool
trampo_OPObjectPane_X (OPObject* obj);

OPaC_TargetExport
auto_OPObjectPane_X (trampo_OPObjectPane_X,"BrowserInit",OP_RUNTIME(OPObjectPane));

Bool
trampo_OPObjectPane_X (OPObject* obj)
{
	OPObjectPane* _instance = (OPObjectPane*)(obj);
	
	OPObjRootProviderRef provider = OP_NEW (OPObjRootProvider);
	provider->Initialise (OPApp::CurrentApp ());
	OPRect box = _instance->ReturnBox ();
	
	_instance->Initialise (box, provider);
	return TRUE;
}

/****************************************************************/

OPObjectPane::OPObjectPane ()
{
	OP_METHOD (OPObjectPane, 0);
	
	this->mode = 0;
	
	this->large_icon_dx = 48;
	this->large_icon_dy = 48;
	this->small_icon_dx = 16;
	this->small_icon_dy = 16;
	
	this->name_dx = 100;
	this->name_dy = 20;
	
	this->cell_dx = 0;
	this->cell_dy = 0;
	
	this->num_objects = 0;
	this->num_columns = 0;
	this->num_lines   = 0;
}

OPObjectPane::~OPObjectPane ()
{
}

void
OPObjectPane::Delete ()
{
	OP_METHOD (Delete, 0);
	
	this->provider = OPObjectNull;
	
	super::Delete ();
}

void
OPObjectPane::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->mode;
	storage << this->provider;
	storage << OPStorageSync::mark;
}

void
OPObjectPane::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->mode;
	storage >> this->provider;
	storage >> OPStorageSync::mark;
}



void
OPObjectPane::Initialise (OPCellGridIn grid, OPRectIn box)
{
	OP_METHOD (Initialise, 2);
	OP_ARG (1, OPCellGridIn, grid);
	OP_ARG (2, OPRectIn, box);
	
	OP_ABORT ("Invalid use of OPObjectPane::Initialise");
}

void
OPObjectPane::Initialise (OPRectIn box, OPGenericProviderIn provider)
{
	OP_METHOD (Initialise, 2);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPGenericProviderIn, provider);
	
	this->Initialise (box, provider, "OPCellGrid");
}

void
OPObjectPane::Initialise (OPRectIn box, OPGenericProviderIn provider, OPAtomIn model)
{
	OP_METHOD (Initialise, 3);
	OP_ARG (1, OPRectIn, box);
	OP_ARG (2, OPGenericProviderIn, provider);
	OP_ARG (3, OPAtomIn, model);
	
	OPCellGridRef grid = OP_DYNAMIC_NEW (model, OPCellGrid);
	
	super::Initialise (grid, box);
	
	this->provider = provider;
	this->Update ();
}

Bool
OPObjectPane::SetMode (OPObjectPaneMode mode)
{
	OP_METHOD (SetMode, 1);
	OP_ARG (1, OPObjectPaneMode, mode);
	
	this->mode = mode;
	this->Update ();
	
	return TRUE;
}

Bool
OPObjectPane::GetMode (OPObjectPaneModeOut mode) const
{
	OP_METHOD (GetMode, 1);
	OP_ARG (1, OPObjectPaneModeOut, mode);
	
	mode = this->mode;
	
	return TRUE;
}

OPObjectPaneMode
OPObjectPane::ReturnMode () const
{
	OP_METHOD (ReturnMode, 0);
	return this->mode;
}


Bool
OPObjectPane::SetProvider (OPGenericProviderIn provider)
{
	OP_METHOD (SetProvider, 1);
	OP_ARG (1, OPGenericProviderIn, provider);
	
	this->provider = provider;
	this->Update ();
	
	return TRUE;
}

Bool
OPObjectPane::GetProvider (OPGenericProviderOut provider) const
{
	OP_METHOD (GetProvider, 1);
	OP_ARG (1, OPGenericProviderOut, provider);
	
	provider = this->provider;
	
	return TRUE;
}


/*
 *	Remove every object which had been added to the
 *	cell grid.
 */

void
OPObjectPane::CleanUp ()
{
	OP_METHOD (CleanUp, 0);
	
	OPControlRef cell;
	
	//	Remove every cell in the grid...
	
	for (Count x = 0; x < this->num_columns; x++) {
		for (Count y = 0; y < this->num_lines; y++) {
			if (this->grid->GetCell (x, y, cell)) {
				this->grid->SetCell (x, y, 0);
				cell->Delete ();
			}
		}
	}
}


/*
 *	Define the size of the name.
 */

Bool
OPObjectPane::SetNameDX (OPCoord dx)
{
	OP_METHOD (SetNameDX, 1);
	OP_ARG (1, OPCoord, dx);
	
	this->name_dx = dx;
	
	return TRUE;
}

Bool
OPObjectPane::SetNameDY (OPCoord dy)
{
	OP_METHOD (SetNameDY, 1);
	OP_ARG (1, OPCoord, dy);
	
	this->name_dy = dy;
	
	return TRUE;
}

OPCoord
OPObjectPane::ReturnNameDX () const
{
	OP_METHOD (ReturnNameDX, 0);
	return this->name_dx;
}

OPCoord
OPObjectPane::ReturnNameDY () const
{
	OP_METHOD (ReturnNameDY, 0);
	return this->name_dy;
}


/*
 *	Find which size will be used by each cell.
 */

void
OPObjectPane::ComputeCellSize ()
{
	OP_METHOD (ComputeCellSize, 0);
	
	OPObjectPaneMode mode = this->mode & OP_PANE_MODE_MASK;
	
	OPCoord dx = 0;
	OPCoord dy = 0;
	
	switch (mode) {
		
		case OP_PANE_MODE_ICON_LARGE:
			this->cell_dx = this->large_icon_dx > this->name_dx ? this->large_icon_dx : this->name_dx;
			this->cell_dy = this->large_icon_dy + this->name_dy;
			break;
		
		case OP_PANE_MODE_ICON_SMALL:
			this->cell_dx = this->small_icon_dx + 2 + this->name_dx;
			this->cell_dy = this->small_icon_dy > this->name_dy ? this->small_icon_dy : this->name_dy;
			
			this->GetSize (dx, dy); dx -= 2; dy -= 2;
			
			this->cell_dx = dx;
			
			if (this->num_objects * this->cell_dy > dy) {
				this->cell_dx -= 17;
			}
			break;
		
		case OP_PANE_MODE_NAME_ONLY:
			this->cell_dx = this->name_dx;
			this->cell_dy = this->name_dy;
			
			this->GetSize (dx, dy); dx -= 2; dy -= 2;
			
			this->cell_dx = dx;
			
			if (this->num_objects * this->cell_dy > dy) {
				this->cell_dx -= 17;
			}
			break;
	}
}


/*
 *	Find how many lines and columns have to be used in order
 *	to get every object in the grid.
 */

void
OPObjectPane::ComputeLayout ()
{
	OP_METHOD (ComputeLayout, 0);
	
	OPCoord dx;
	OPCoord dy;
	
	this->num_objects = (this->provider.IsValid ()) ? this->provider->ReturnNumObjects () : 0;
	this->scroll_view->GetSize (dx, dy);
	
	OPObjectPaneMode mode = this->mode & OP_PANE_LAYOUT_MASK;
	
	OP_ASSERT (this->cell_dx > 0);
	OP_ASSERT (this->cell_dy > 0);
	
	switch (mode) {
		
		case OP_PANE_LAYOUT_LINES:
			this->num_columns = (Count)(dx / this->cell_dx);
			if (this->num_columns == 0) this->num_columns = 1;
			this->num_lines = (this->num_objects + this->num_columns - 1) / this->num_columns;
			if (this->num_lines == 0) this->num_lines = 1;
			break;
		
		case OP_PANE_LAYOUT_COLUMNS:
			this->num_lines = (Count)(dy / this->cell_dy);
			if (this->num_lines == 0) this->num_lines = 1;
			this->num_columns = (this->num_objects + this->num_lines - 1) / this->num_lines;
			if (this->num_columns == 0) this->num_columns = 1;
			break;
	}
}



/*
 *	The size of either the cell grid or the scroll grid changed.
 *	We might want to change the pane layout...
 */

void
OPObjectPane::GridChangedSize ()
{
	OP_METHOD (GridChangedSize, 0);
	
	this->ComputeCellSize ();
	this->ComputeLayout ();

	super::GridChangedSize ();
	
	this->Update ();
}



/*
 *	Find the position (column and line) in the grid for the
 *	specified cell.
 */

Bool
OPObjectPane::GetCellLayout (Count index, CountOut x, CountOut y) const
{
	OP_METHOD (GetCellLayout, 3);
	OP_ARG (1, Count, index);
	OP_ARG (2, CountOut, x);
	OP_ARG (3, CountOut, y);
	
	OPObjectPaneMode mode = this->mode & OP_PANE_LAYOUT_MASK;
	
	if (mode == OP_PANE_LAYOUT_LINES) {
		
		x = index % this->num_columns;
		y = index / this->num_columns;
		
	} else {
		
		x = index / this->num_lines;
		y = index % this->num_lines;
	}
	
	return TRUE;
}


/*
 *	Convert from position to co-ordinates and vice versa.
 */

Bool
OPObjectPane::GetRankFromPos (Count x, Count y, CountOut rank) const
{
	OP_METHOD (GetRankFromPos, 3);
	OP_ARG (1, Count, x);
	OP_ARG (2, Count, y);
	OP_ARG (3, CountOut, rank);
	
	if ( (x >= this->num_columns)
	  || (y >= this->num_lines) ) {
		return FALSE;
	}
	
	OPObjectPaneMode mode = this->mode & OP_PANE_LAYOUT_MASK;
	
	if (mode == OP_PANE_LAYOUT_LINES) {
		
		rank = y * this->num_columns + x;
		
	} else {
		
		rank = x * this->num_lines + y;
	}
	
	return TRUE;
}

Bool
OPObjectPane::GetPosFromRank (Count rank, CountOut x, CountOut y) const
{
	OP_METHOD (GetPosFromRank, 3);
	OP_ARG (1, Count, rank);
	OP_ARG (2, CountOut, x);
	OP_ARG (3, CountOut, y);
	
	return this->GetCellLayout (rank, x, y);
}


/*
 *	Create a cell with an icon and a name.
 */

Bool
OPObjectPane::CreateCell (Count index, OPControlOut cell)
{
	OP_METHOD (CreateCell, 2);
	OP_ARG (1, Count, index);
	OP_ARG (2, OPControlOut, cell);
	
	if (this->provider.IsInvalid ()) {
		return FALSE;
	}
	
	OPObjectPaneMode i_mode = this->mode & OP_PANE_MODE_MASK;
	OPLookIcon look_icon = OP_ICON_NONE;
	OPStringRef text;
	OPObjectRef object;
	OPCoord x = 0;
	OPCoord y = 0;
	
	OPBoxViewRef box  = OP_NEW (OPBoxView);
	OPIconRef    icon = OP_NEW (OPIcon);
	OPTextTagRef tag  = OP_NEW (OPTextTag);
	
	box->SetSize (this->cell_dx, this->cell_dy);
	box->EnableTitle (FALSE);
	box->EnableFrame (FALSE);
	
	this->provider->GetObject (index, object);
	
	switch (i_mode) {
		
		case OP_PANE_MODE_ICON_LARGE:
			
			x = (this->cell_dx - this->large_icon_dx) / 2;
			y = this->large_icon_dy;
			
			this->provider->GetLargeIcon (index, look_icon);
			this->provider->GetText (index, text);
			icon->Initialise (look_icon);
			icon->SetParent (box);
			icon->SetOrigin (x, 0);
			icon->SetSize (this->large_icon_dx, this->large_icon_dy);
			icon->SetFlagOkDetect (FALSE);
			tag->SetTitle (text);
			tag->SetParent (box);
			tag->SetOrigin (0, y);
			tag->SetSize (this->name_dx, this->name_dy);
			break;
		
		case OP_PANE_MODE_ICON_SMALL:
			
			x = this->small_icon_dx+2;
			y = (this->cell_dy - this->small_icon_dy) / 2;
			
			this->provider->GetSmallIcon (index, look_icon);
			this->provider->GetText (index, text);
			icon->Initialise (look_icon);
			icon->SetParent (box);
			icon->SetOrigin (2, 0);
			icon->SetSize (this->small_icon_dx, this->cell_dy);
			icon->SetFlagOkDetect (FALSE);
			tag->SetTitle (text);
			tag->SetParent (box);
			tag->SetOrigin (x, 0);
			tag->SetMode (OP_TEXT_HPOS_LEFT | OP_TEXT_VPOS_CENTER);
			tag->SetSize (this->cell_dx-x, this->cell_dy);
			tag->SetLeftSpace (4);
			break;
			
		case OP_PANE_MODE_NAME_ONLY:
			
			this->provider->GetText (index, text);
			tag->SetTitle (text);
			tag->SetParent (box);
			tag->SetOrigin (x, 0);
			tag->SetSize (this->cell_dx, this->cell_dy);
			tag->SetMode (OP_TEXT_HPOS_LEFT | OP_TEXT_VPOS_CENTER);
			tag->SetLeftSpace (4);
			break;
	}
	
	if (this->provider->GetMarkIcon (index, look_icon)) {
		
		OPCoord dx, dy, mdx, mdy;
		
		icon = OP_NEW (OPIcon);
		icon->Initialise (look_icon);
		icon->SetParent (box);
		icon->GetMinSize (mdx, mdy);
		tag->GetOrigin (x, y);
		tag->GetSize (dx, dy);
		icon->SetOrigin (x+dx-mdx-2, y);
		icon->SetSize (mdx, dy);
		tag->SetSize (dx-mdx-4, dy);
	}
	
	cell = box;
	
	cell->SetHiliteMode (TRUE);
	cell->SetShapeStyle (cell->ReturnShapeStyle () | OP_CTRL_AUTOHILITE | OP_CTRL_AUTOACTHILITE);
	cell->SetAttribute ("EmbeddedObjectForConnect", object);
	
	return TRUE;
}


/*
 *	Update the contents of the pane by asking the provider
 *	about the icons and names for the objects.
 */

Bool
OPObjectPane::Update ()
{
	OP_METHOD (Update, 0);
	
	OPSelGridRef sel;
	
	Count  num_sel   = 0;
	Count* sel_cells = 0;
	
	//	Remember about the selected cells... if the cell grid is a selection
	//	grid. Otherwise, just rebuild the grid without worrying about the cell's
	//	state.
	
	if ( (this->grid.IsValid ())
	  && (this->grid->IsKindOf (OP_RUNTIME (OPSelGrid))) ) {
		
		sel       = this->grid;
		num_sel   = sel->ReturnSelectNum ();
		sel_cells = OP_ALLOC_ARRAY (Count, 2*num_sel);

		for (Count index = 0; index < num_sel; index++) {
			sel->FindActiveHilite (sel_cells[index*2], sel_cells[index*2+1], index);
		}
	}
	
	this->CleanUp ();	
	this->num_objects = (this->provider.IsValid ()) ? this->provider->ReturnNumObjects () : 0;
	
	this->ComputeCellSize ();
	this->ComputeLayout ();
	
	this->grid->SetArraySize (this->num_columns, this->num_lines);
	this->grid->SetIsometricMode (TRUE);
	this->grid->SetCellSize (this->cell_dx, this->cell_dy);
	this->grid->UpdateGridGeometry ();
	
	Count x, y;
	OPControlRef cell;
	
	for (Count index = 0; index < this->num_objects; index++) {
		this->CreateCell (index, cell);
		this->GetCellLayout (index, x, y);
		this->grid->SetCell (x, y, cell);
	}
	
	//	If there were any selected cells, select them back again.
	
	if (num_sel) {
		
		for (Count index = 0; index < num_sel; index++) {
			sel->SetActiveHilite (sel_cells[index*2], sel_cells[index*2+1], TRUE, FALSE);
		}
		
		OP_FREE (sel_cells);
	}
	
	return TRUE;
}

