/*
 *	spring.cxx
 *
 *	A spring maps a stretching factor with a size and can be used to
 *	dynamically resize and move objects within a cell.
 *
 *	(C) Copyright 1994-1997, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/new.h>
#include <OPaC/spring.h>
#include <stdlib.h>

/****************************************************************/

OP_IMPLEMENTATION (spring.h);
OP_IMPLEMENT_CLASS (OPSpring);

/****************************************************************/

/*
 *	Create an empty spring curve. Use Reset to specify a default
 *	size.
 */

OPSpring::OPSpring ()
{
	OP_METHOD (OPSpring, 0);

	this->head  = 0;
	this->tail  = 0;
	this->count = 0;

	this->Reset (0);
}

/*
 *	Destroy and free memory.
 */

OPSpring::~OPSpring ()
{
	OP_METHOD (~OPSpring, 0);

	this->RemoveAll ();
}

/*
 *	Store the spring curve.
 */
	
void
OPSpring::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->count;
	storage << OPStorageSync::mark;

	OPSpringSample* sample = this->head;
	
	while (sample) {
		storage << sample->stretch;
		storage << sample->size;
		storage << OPStorageSync::mark;
		sample = sample->next;
	}
}



/*
 *	Restore the spring curve.
 */

void
OPSpring::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	this->RemoveAll ();

	Count count = 0;
	storage >> count;
	storage >> OPStorageSync::mark;
	
	while (count--) {
		Real    stretch = 0;
		OPCoord size    = 0;
		storage >> stretch;
		storage >> size;
		storage >> OPStorageSync::mark;
		this->Insert (stretch, size);
	}
}



/*
 *	Prepare a spring curve for its destruction. Nothing will
 *	will be done; it is up to the destructor to free memory.
 */

void
OPSpring::Delete ()
{
	OP_METHOD (Delete, 0);

	super::Delete ();
}



/*
 *	Remove every sample from the spring curve.
 */

void
OPSpring::RemoveAll ()
{
	OP_METHOD (RemoveAll, 0);

	OPSpringSample* sample = this->head;
	OPSpringSample* next   = 0;
	
	while (sample) {
		next   = sample->next;
		OP_FREE (sample);
		sample = next;
	}
	
	this->head  = 0;
	this->tail  = 0;
	this->count = 0;
}



/*
 *	Reset the spring curve to a constant between the minimal and
 *	maximal stretch values.
 */

void
OPSpring::Reset (OPCoord size)
{
	OP_METHOD (Reset, 1);
	OP_ARG (1, OPCoord, size);

	this->RemoveAll ();
	this->Insert (OP_SPRING_MIN, size);
	this->Insert (OP_SPRING_MAX, size);
}


/*
 *	Merge the given spring band with this one. It is as if
 *	both springs were soldered together.
 */

void
OPSpring::Merge (OPSpringIn spring)
{
	OP_METHOD (Merge, 1);
	OP_ARG (1, OPSpringIn, spring);

	Count num   = 0;
	Count total = 0;
	
	OPSpringSample* sample = spring->head;
	OPSpringSample* head   = 0;
	
	//	First walk through the inserted spring knots.
	
	num = spring->count;
	
	while (num--) {
		
		Real    stretch = sample->stretch;
		OPCoord size    = sample->size + this->FindSize (stretch);
		
		OPSpringSample* n = (OPSpringSample*)alloca (sizeof (OPSpringSample));
		
		n->next    = head;
		n->stretch = stretch;
		n->size    = size;
		
		head   = n;
		sample = sample->next;
		total++;
	}

	num    = this->count;
	sample = this->head;
	
	while (num--) {
		
		Real    stretch = sample->stretch;
		OPCoord size    = sample->size + spring->FindSize (stretch);
		
		OPSpringSample* n = (OPSpringSample*)alloca (sizeof (OPSpringSample));
		
		n->next    = head;
		n->stretch = stretch;
		n->size    = size;
		
		head   = n;
		sample = sample->next;
		total++;
	}
	
	this->Reset (0);
	num = total;

	while (num--) {
		this->Insert (head->stretch, head->size);
		head = head->next;
	}
}


/*
 *	Insert just one point into the spring curve. If it already
 *	exists, replace the previous one.
 *
 *	Just reject samples which would lie out of bounds.
 */

void
OPSpring::Insert (Real stretch, OPCoord size)
{
	OP_METHOD (Insert, 2);
	OP_ARG (1, Real, stretch);
	OP_ARG (2, OPCoord, size);

	if ( (stretch > OP_SPRING_MAX)
	  || (stretch < OP_SPRING_MIN) ) {
		return;
	}
	
	OPSpringSample* sample = this->head;
	OPSpringSample* prev   = 0;

	//	Optimize insertions at the end of the curve.
	
	if ( (this->count > 1)
	  && (stretch < this->tail->stretch)
	  && (stretch > this->tail->prev->stretch) ) {
		sample = this->tail;
		prev   = sample->prev;
		goto insert;
	}
	
	while (sample) {
		
		//	If we have found a sample with the same stretch value, just
		//	replace it.
		
		if (sample->stretch == stretch) {
			sample->size = size;
			return;
		}
		
		//	If we have moved too far, stop and insert after the previous.
		
		if (stretch < sample->stretch) {
			break;
		}
		
		prev   = sample;
		sample = sample->next;
	}

insert:
	sample = OP_ALLOC (OPSpringSample);

	sample->next    = (prev) ? (prev->next) : this->head;
	sample->prev    = prev;
	sample->stretch = stretch;
	sample->size    = size;
	
	(sample->next ? sample->next->prev : this->tail) = sample;
	(sample->prev ? sample->prev->next : this->head) = sample;

	this->count++;
}



/*
 *	Find the specified sample and remove it if it exists.
 *	You may not remove the minimal and maximal samples.
 */

void
OPSpring::Remove (Real stretch)
{
	OP_METHOD (Remove, 1);
	OP_ARG (1, Real, stretch);
	
	if ( (stretch > OP_SPRING_MAX)
	  || (stretch < OP_SPRING_MIN)
	  || (this->head == 0) ) {
		return;
	}
	
	//	Special case : want to remove minimum or maximum, so just replace
	//	them to produce a horizontal first or last segment in the curve.
	
	if (stretch == OP_SPRING_MIN) {
		OP_ASSERT (this->head->next);
		this->head->stretch = this->head->next->stretch;
		return;
	}

	if (stretch == OP_SPRING_MAX) {
		OP_ASSERT (this->tail->prev);
		this->tail->stretch = this->tail->prev->stretch;
		return;
	}
	
	OPSpringSample* sample = this->head;

	while (sample) {
		
		if (sample->stretch == stretch) {
			
			//	Remove the sample from the list.
			
			(sample->next ? sample->next->prev : this->tail) = sample->next;
			(sample->prev ? sample->prev->next : this->head) = sample->prev;
			
			this->count--;
			
			OP_FREE (sample);
			return;
		}
		
		sample = sample->next;
	}
}



/*
 *	Return the size specified for the given stretch factor.
 *	A linear interpolation is given.
 */

OPCoord
OPSpring::FindSize (Real stretch) const
{
	OP_METHOD (FindSize, 1);
	OP_ARG (1, Real, stretch);

	if (stretch < OP_SPRING_MIN) {
		stretch = OP_SPRING_MIN;
	}

	if (stretch > OP_SPRING_MAX) {
		stretch = OP_SPRING_MAX;
	}

	OPSpringSample* sample = this->head;
	OPSpringSample* prev   = 0;

	if (sample == 0) {
		return 0;
	}
	
	while (sample) {
		
		if (sample->stretch == stretch) {
			return sample->size;
		}
		
		if (stretch < sample->stretch) {
			break;
		}
		
		prev   = sample;
		sample = sample->next;
	}

	OP_ASSERT (sample);
	OP_ASSERT (prev);
	
	Real d_stretch = sample->stretch - prev->stretch;			// never 0
	Real d_size    = sample->size    - prev->size;

	stretch -= prev->stretch;

	return (OPCoord)(prev->size + d_size * stretch / d_stretch);
}


/*
 *	For any given size, find the stretch factor that fits
 *	best.
 */

Real
OPSpring::FindStretch (OPCoord size) const
{
	OP_METHOD (FindStretch, 1);
	OP_ARG (1, OPCoord, size);
	
	Real stretch = 0;

	OPSpringSample* sample = this->head;
	OPSpringSample* next   = 0;
	
	if (sample == 0) {
		return OP_SPRING_ONE;
	}
	
	if (sample->size == size) {
		return sample->stretch;
	}
	
	while ( (sample) && (sample->next) ) {
		
		next = sample->next;
		
		if (next->size == size) {
			return next->stretch;
		}
		
		if ( (sample->size < size)
		  && (next->size > size) ) {
			Real d_stretch = next->stretch - sample->stretch;
			Real d_size    = next->size    - sample->size;		// never 0
			
			size   -= sample->size;
			stretch = sample->stretch + d_stretch * size / d_size;
			
			return stretch;
		}
		
		if ( (next->size < size)
		  && (sample->size > size) ) {
			Real d_stretch = sample->stretch - next->stretch;
			Real d_size    = sample->size    - next->size;		// never 0
			
			size   -= next->size;
			stretch = next->stretch + d_stretch * size / d_size;
			
			return stretch;
		}
		
		sample = next;
	}

	if (size > this->tail->size) {
		return this->tail->stretch;
	}
	
	if (size < this->head->size) {
		return this->head->stretch;
	}
	
	return OP_SPRING_ONE;
}



/*
 *	Return number of samples.
 */

Count
OPSpring::CountSamples () const
{
	OP_METHOD (CountSamples, 0);
	
	return this->count;
}



/*
 *	Return the description of the nth sample. TRUE if the
 *	sample exists.
 */

Bool
OPSpring::GetNthSample (Count index, RealOut stretch, OPCoordOut size)
{
	OP_METHOD (GetNthSample, 3);
	OP_ARG (1, Count, index);
	OP_ARG (2, RealOut, stretch);
	OP_ARG (3, OPCoordOut, size);

	OPSpringSample* sample = this->head;
	
	stretch = 0;
	size    = 0;

	if (index >= this->count) {
		return FALSE;
	}

	while (index--) {
		sample = sample->next;
	}

	stretch = sample->stretch;
	size    = sample->size;
	
	return TRUE;
}

