/*
 *	tablist.cxx
 *
 *	List of tabs for a tabbed view. This is rarely used directly when
 *	you implement tabbed views. Use OPTabView instead.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/tablist.h>
#include <OPaC/lookman.h>
#include <OPaC/container.h>
#include <OPaC/grail.h>
#include <OPaC/window.h>
#include <OPaC/event.h>
#include <OPaC/pubmacro.h>
#include <OPaC/notmacro.h>

/****************************************************************/

OP_IMPLEMENTATION (tablist.h);
OP_IMPLEMENT_CLASS (OPTabList);

/****************************************************************/

OP_NOTIFY_EVENT (OPTabList, OnTabChange);

/****************************************************************/

OP_PUBLISH_MODIFUNC (OPTabList, "ActiveTab", ReturnActiveTab, SetActiveTab, "Active tab", Card32, Card32);
OP_PUBLISH_MODIFUNC (OPTabList, "FirstTab",  ReturnFirstTab,  SetFirstTab,  "First tab",  Card32, Card32);

/****************************************************************/

/*
 *	Initialise an empty tabbed list.
 */

OPTabList::OPTabList ()
{
	OP_METHOD (OPTabList, 0);

	this->space  = 2;
	this->first  = 0;
	this->active = 0;
	
	this->need_update = TRUE;
}


/*
 *	Destroy the tabbed list. If we still have some tabbed texts
 *	tell them that we no longer exist.
 */

OPTabList::~OPTabList ()
{
	OP_METHOD (~OPTabList, 0);

	if (this->tabs.IsValid ()) {
		this->SetTabsParent (OPViewNull);
	}
}


/*
 *	Prepare the tabbed view for its destruction. Remove any
 *	links to other objects, such as the tabbed texts.
 */

void
OPTabList::Delete ()
{
	OP_METHOD (Delete, 0);
	
	OPViewRef none;

	this->SetTabsParent (none);
	
	if (this->tabs.IsValid ()) {
		this->tabs->Delete ();
	}
	
	this->tabs        = OPListNull; 
	this->need_update = TRUE;
	
	super::Delete ();
}


/*
 *	Overridden to take into account the subviews stored in
 *	the tabs, which are not considered to be normal child
 *	views.
 */

Bool
OPTabList::SetWindow (OPWindowIn window)
{
	OP_METHOD (SetWindow, 1);
	OP_ARG (1, OPWindowIn, window);
	
	//	If the window which has been specified is equal to the current window,
	//	don't do anything...
	
	if (this->window == window) {
		return TRUE;
	}
	
	if (super::SetWindow (window)) {
		this->SetTabsParent (this);
		return TRUE;
	}

	return FALSE;
}



/*
 *	Change the window of the subviews stored in the tabs,
 *	depending on the parent (this or none).
 */

void
OPTabList::SetTabsParent (OPViewIn view)
{
	OP_METHOD (SetTabsParent, 1);
	OP_ARG (1, OPViewIn, view);
	
	OPWindowRef window;

	if (view.IsValid ()) {
		view->GetWindow (window);
	}
	
	if (this->tabs.IsValid ()) {
		
		Count index;
		Count max = this->tabs->TotalCount ();
		
		for (index = 0; index < max; index++) {
			
			OPObjectRef object;
			OPViewRef   child;
			
			this->tabs->Get (index, object);
			
			child = object;
			child->SetWindow (window);
		}
	}
}


/*
 *	Store the tabs.
 */

void
OPTabList::StoreInstance (OPStorageIn storage)
{
	OP_METHOD (StoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::StoreInstance (storage);
	
	storage << this->space;
	storage << this->first << this->active;
	storage << this->tabs;
	storage << OPStorageSync::mark;
}


/*
 *	Restore the tabs.
 */

void
OPTabList::RestoreInstance (OPStorageIn storage)
{
	OP_METHOD (RestoreInstance, 1);
	OP_ARG (1, OPStorageIn, storage);
	
	super::RestoreInstance (storage);
	
	storage >> this->space;
	storage >> this->first >> this->active;
	storage >> this->tabs;
	storage >> OPStorageSync::mark;
	
	this->need_update = TRUE;
}



void
OPTabList::ActionOn ()
{
	super::ActionOn ();
}

void
OPTabList::ActionOff ()
{
	super::ActionOff ();
}

void
OPTabList::Action (Bool clicked)
{
	super::Action (clicked);
}



/*
 *	Find in which tab the pointer currently is. 0 => none.
 */

Count
OPTabList::FindTab (OPEventIn event)
{
	OP_METHOD (FindTab, 1);
	OP_ARG (1, OPEventIn, event);
	
	Count max   = this->tabs->TotalCount ();
	Count index = 0;
	
	OPCoord x = event->ReturnX ();
	OPCoord y = event->ReturnY ();
	
	for (index = this->first; index <= max; index++) {
		
		OPObjectRef object;
		OPViewRef   view;
		OPRect      box;
		
		this->tabs->Get (index-1, object);
		
		view = object;
		
		box = view->ReturnBox ();

		box.SetSize (box.DX ()+this->space*2, box.DY ()+this->space*2);
		box.SetOrigin (box.X1 ()-this->space, box.Y1 ()-this->space);
		
		if (box.IsInside (x, y)) {
			return index;
		}
	}
	
	return 0;
}


/*
 *	Compute widths of the miscellanous tabs.
 */

void
OPTabList::ComputeWidths (OPCoord width[])
{
	Count max   = this->tabs->TotalCount ();
	Count index = 0;
	
	for (index = 0; index < max; index++) {
		
		OPObjectRef object;
		OPViewRef   view;
		OPCoord     dx = 0;
		OPCoord     dy = 0;
		
		this->tabs->Get (index, object);
		
		view = object;
		view->GetMinSize (dx, dy);
		view->SetSize (dx, dy);
		
		width[index] = dx+this->space+8;
	}
}


/*
 *	Compute the geometry and update information about the
 *	first visible tab in the list.
 */

void
OPTabList::UpdateTabs ()
{
	OP_METHOD (UpdateTabs, 0);
	OP_ASSERT (this->tabs.IsValid ());
	
	Count max   = this->tabs->TotalCount ();
	Count index = 0;
	
	OP_ASSERT (max < 100);

	OPCoord width[100];
	
	this->ComputeWidths (width);

	for (;;) {
		
		if ( (this->first == this->active)
		  || (this->active == 0) ) {
			break;
		}
		
		if ( (this->first > max)
		  || (this->first == 0) ) {
			this->first = 1;
		}
		
		OPCoord to_end   = 0;
		OPCoord to_begin = 0;
		OPCoord offset   = 0;

		for (index = 1; index < this->first; index++) {
			to_begin += width[index-1];
		}

		for (index = this->first; index <= max; index++) {
			to_end += width[index-1];
		}

		for (index = 1; index <= this->active; index++) {
			offset += width[index-1];
		}

		//	The tab which should be shown is fully visible.
		
		if ( (offset >= to_begin)
		  && (offset-to_begin+width[this->active-1] <= this->box.DX ()) ) {
			break;
		}
		
		//	The tab which should be shown is not visible, since the
		//	first tab is on the right of it. So move it left.
		
		if (offset < to_begin) {
			this->first--;
			continue;
		}
		
		//	The tab which should be shown is not (fully) visible,
		//	since it is placed too far on the right.
		
		if (offset-to_begin+width[this->active-1] > this->box.DX ()) {
			this->first++;
			continue;
		}
	}

	this->need_update = FALSE;
}


/*
 *	Draw just one tab at the specified coordinate. If the
 *	tab is enlarged, it is on top of the others.
 */

void
OPTabList::DrawTab (OPCoord x, OPCoord y, OPCoord dx,
					OPGrailPortIn port, OPViewIn view, Bool enlarged)
{
	OP_METHOD (DrawTab, 6);
	OP_ARG (1, OPCoord, x);
	OP_ARG (2, OPCoord, y);
	OP_ARG (3, OPCoord, dx);
	OP_ARG (4, OPGrailPortIn, port);
	OP_ARG (5, OPViewIn, view);
	OP_ARG (6, Bool, enlarged);
	
	const OPAdornerRef& adorner = port->ReturnAdorner ();
	
	OPCoord dy = this->box.DY ()-4;
	OPCoord ox = x+this->space+3;

	if (enlarged) {
		y  += this->space;
		dx += 2*this->space;
		dy += this->space;
		ox += this->space;
	} else {
		x  += this->space;
		dy -= this->space;
	}

	adorner->SetForeColor (OP_COLOR_LGRAY);
	port->DrawLineDY (x-1, y-dy+2, dy-3);
	port->DrawLineDX (x+1, y, dx-3);
	port->DrawLineDX (x+1, y-1, dx-3);
	
	adorner->SetForeColor (OP_COLOR_WHITE);
	port->DrawLineDY (x, y-dy+2, dy-2);
	port->DrawLineDY (x+1, y-dy+1, 1);
	port->DrawLineDX (x+2, y-dy, dx-4);
	
	adorner->SetForeColor (OP_COLOR_DGRAY);
	port->DrawLineDY (x+dx-2, y-dy, dy);

	adorner->SetForeColor (OP_COLOR_BLACK);
	port->DrawLineDY (x+dx-1, y-dy+2, dy-2);

	view->SetOrigin (ox, y-dy+3);
	view->Draw (port, this->box);
}


/*
 *	Define the list of tabs. The tabs should be views (eg.
 *	text tags or icons, etc.)
 */

Bool 
OPTabList::SetTabs (OPListIn tabs)
{
	OP_METHOD (SetTabs, 1);
	OP_ARG (1, OPListIn, tabs);
	
	if (this->tabs.IsValid ()) {
		OPViewRef none;
		this->SetTabsParent (none);
	}
	
	this->tabs        = tabs;
	this->need_update = TRUE;
	
	this->SetTabsParent (this);
	
	return TRUE;
}


/*
 *	Define the space between the tabs. Default is 2.
 */

Bool 
OPTabList::SetSpace (OPCoord space)
{
	OP_METHOD (SetSpace, 1);
	OP_ARG (1, OPCoord, space);

	this->space       = space;
	this->need_update = TRUE;
	
	return TRUE;
}


/*
 *	Define the first visible tab in the list. If this
 *	will make the active tab invisible, it will be
 *	adjusted.
 */

Bool 
OPTabList::SetFirstTab (Count first)
{
	OP_METHOD (SetFirstTab, 1);
	OP_ARG (1, Count, first);

	this->first       = first;
	this->need_update = TRUE;
	return TRUE;
}


/*
 *	Define the active tab (the one on top of the others).
 *	This takes precedence over the first visible tab.
 *	If active is set to 0, no tab will be active.
 */

Bool 
OPTabList::SetActiveTab (Count active)
{
	OP_METHOD (SetActiveTab, 1);
	OP_ARG (1, Count, active);

	if (active == this->active) {
		return TRUE;
	}
	
	this->active      = active;
	this->need_update = TRUE;
	
	this->Report (notify_OnTabChange, OP_REPORT_NEW);
	
	return TRUE;
}


/*
 *	Return information about the tabs.
 */

Bool    
OPTabList::GetTabs (OPListOut tabs) const
{
	OP_METHOD (GetTabs, 1);
	OP_ARG (1, OPListOut, tabs);

	tabs = this->tabs;
	return TRUE;
}

Count  
OPTabList::CountTabs () const
{
	OP_METHOD (CountTabs, 0);
	return this->tabs.IsValid () ? this->tabs->TotalCount () : 0;
}

/*
 *	Return the space between the tabs.
 */

OPCoord 
OPTabList::ReturnSpace () const
{
	OP_METHOD (ReturnSpace, 0);
	
	return this->space;
}


/*
 *	Return the first visible tab (1..n).
 */

Count   
OPTabList::ReturnFirstTab () const
{
	OP_METHOD (ReturnFirstTab, 0);

	if (this->need_update) {
		OPTabList* that = const_cast<OPTabList*> (this);
		that->UpdateTabs ();
	}
	
	return this->first;
}


/*
 *	Return the active tab number (0..n). 0 => none.
 */

Count   
OPTabList::ReturnActiveTab () const
{
	OP_METHOD (ReturnActiveTab, 0);

	return this->active;
}


/*
 *	Draw the tabs.
 */

void
OPTabList::Draw (OPGrailPortIn port, OPRectIn rect, OPDrawOp what)
{
	OP_METHOD (Draw, 3);
	OP_ARG (1, OPGrailPortIn, port);
	OP_ARG (2, OPRectIn, rect);
	OP_ARG (3, OPDrawOp, what);
	
	super::Draw (port, rect, what);
	
	if (what & OP_DRAW_OBJECT) {
		
		if (this->need_update) {
			this->UpdateTabs ();
		}
		
		if (this->first) {
			
			Count   max    = this->tabs->TotalCount ();
			Count   index  = 0;
			OPCoord x      = this->box.X1 ();
			OPCoord y      = this->box.Y2 () - 2;
			OPCoord hit_x  = 0;
			OPCoord hit_dx = 0;
			
			for (index = this->first; index <= max; index++) {
				
				OPObjectRef object;
				OPViewRef   view;
				OPCoord     dx = 0;
				OPCoord     dy = 0;
				
				this->tabs->Get (index-1, object);
				
				view = object;
				view->GetMinSize (dx, dy);
				
				dx += 2*4;
				
				if (index == this->active) {
					hit_x  = x;
					hit_dx = dx;
				} else {
					this->DrawTab (x, y, dx+3, port, view, FALSE);
				}
				
				x += dx;
				x += 2*this->space;
			}

			const OPAdornerRef& adorner = port->ReturnAdorner ();
			adorner->SetForeColor (OP_COLOR_WHITE);
			port->DrawLineDX (this->box.X1 (), this->box.Y2 ()-1, this->box.DX ());

			if (hit_dx != 0) {
				
				OPObjectRef object;
				OPViewRef   view;
				
				this->tabs->Get (this->active-1, object);
				
				view = object;
				
				this->DrawTab (hit_x, y, hit_dx+3, port, view, TRUE);
			}
		}
	}
}

void
OPTabList::ButtonPress (OPEventIn event)
{
	OP_METHOD (ButtonPress, 1);
	OP_ARG (1, OPEventIn, event);

	Count tab = this->FindTab (event);
	
	if ( (tab)
	  && (tab != this->active) ) {
		this->SetActiveTab (tab);
	}
	
	super::ButtonPress (event);
}

void
OPTabList::ButtonRelease (OPEventIn event)
{
	OP_METHOD (ButtonRelease, 1);
	OP_ARG (1, OPEventIn, event);
	
//	super::ButtonRelease (event);
	
	this->window->EnableMainResponder ();
	this->Display ();
}

void
OPTabList::EnterNotify (OPEventIn event)
{
	OP_METHOD (EnterNotify, 1);
	OP_ARG (1, OPEventIn, event);

	super::EnterNotify (event);
}

void
OPTabList::MotionNotify (OPEventIn event)
{
	OP_METHOD (MotionNotify, 1);
	OP_ARG (1, OPEventIn, event);

	super::MotionNotify (event);
}

void
OPTabList::LeaveNotify (OPEventIn event)
{
	OP_METHOD (LeaveNotify, 1);
	OP_ARG (1, OPEventIn, event);

	super::LeaveNotify (event);
}

void
OPTabList::KeyDown (OPEventIn event)
{
	OP_METHOD (KeyDown, 1);
	OP_ARG (1, OPEventIn, event);

	super::KeyDown (event);
}

void
OPTabList::KeyUp (OPEventIn event)
{
	OP_METHOD (KeyUp, 1);
	OP_ARG (1, OPEventIn, event);

	super::KeyUp (event);
}

