/*
 *	gfxsurface.h
 *
 *	Graphic library.
 *
 *	The GfxSurface provides an interface to a true color pixmap. The
 *	pixels are always represented as alpha-RGB (8-8-8-8). They are
 *	converted internally to a more compact format if 8-bit or 16-bit
 *	pixmaps are used.
 *
 *	The  8-bit pixmap is just a monochrome, gray-level pixmap.
 *	The 16-bit pixmap is organised as 5-6-5, which is an RGB standard.
 *	The 32-bit pixmap is organised as alpha-RGB.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#ifndef _GFX_GFXSURFACE_H_
#define	_GFX_GFXSURFACE_H_

#include "gfxlibrary.h"

class GfxFontCache;

/****************************************************************/

/*
 *	The Surface class provides a kind of pixmap interface,
 *	including alpha blending and simple graphic operators.
 */

class GFX_PUBLIC GfxSurface
{
	friend class GfxFontCache;
	
protected:
	union MemoryPtr							//	convenient way to access memory (no casts)
	{
		void*			mem_any;
		Card8*			mem_8;
		Card16*			mem_16;
		Card32*			mem_32;
	};
	
	struct LineDot
	{
		Card32			pos;				//	position of dot (either x or y)
		Card8			error;				//	error from rendering engine
	};
	
public:
	enum ColorPlane
	{
		PLANE_ALPHA		= 0,
		PLANE_RED		= 1,
		PLANE_GREEN		= 2,
		PLANE_BLUE		= 3
	};
	
protected:
	MemoryPtr			base;				//	base memory pointer (if any)
	MemoryPtr			line_buffer;		//	conversion buffer for a single line
	MemoryPtr*			lines;				//	array of line pointers
	LineDot*			dots;				//	array of dots
	Card32				dx;					//	width of the surface, in pixels
	Card32				dy;					//	height of the surface, in pixels
	Card32				pitch;				//	width of the surface, in bytes
	Card32				alpha;				//	global alpha value
	
	Bool				base_allocated;		//	TRUE => allocated base, need to free it
	Bool				is_transparent;		//	TRUE => uses transparent color
	
	Card32				clip_x1, clip_y1;	//	top/left (clipping rectangle)
	Card32				clip_x2, clip_y2;	//	bottom/right (clipping rectangle)
	
	Card32				depth;				//	depth (bits/pixel)
	Card32				pixel_bytes;		//	offset (in bytes) to next pixel
	Card32				pixel_shift;		//	(1 << pixel_shift) == pixel_bytes
	Card32				pixel_trans;		//	transparent pixel (if any)
	
protected:
	virtual void AllocSurface ();
	virtual void AllocLinesForSurface ();
	virtual void AllocDotsBuffer ();
	virtual void FreeSurfaceAndLines ();
	
	void ComputeLineDotsDX (Card32 x, Card32 y, Card32 dx, Int32 dy, Real error = 0);
	void ComputeLineDotsDY (Card32 x, Card32 y, Int32 dx, Card32 dy, Real error = 0);
	
	GfxSurface (Card32 bpp);
	virtual ~ GfxSurface ();
public:
	
	void Initialise (Card32 dx, Card32 dy);
	void Initialise (Card32 dx, Card32 dy, void* memory, Card32 pitch);
	
	Card32 DX () const				{ return this->dx; }
	Card32 DY () const				{ return this->dy; }
	Card32 Depth () const			{ return this->depth; }
	Card32 Alpha () const			{ return this->alpha; }
	Card32 Pitch () const			{ return this->pitch; }
	void* Line (int index) const	{ return this->lines[index].mem_any; }
	
	void SetAlpha (Card32 alpha)	{ this->alpha = alpha; }
	void SetTransparent (Card32 p)	{ this->pixel_trans = p; this->is_transparent = TRUE; }
	void SetOpaque ()				{ this->is_transparent = FALSE; }
	
	void SetClipRect (Card32 x, Card32 y, Card32 dx, Card32 dy);
	void ResetClipRect ();
	
	//	Single pixel manipulation :
	
	void SetPixelRGB (Int32 x, Int32 y, Card32 pixel);
	void SetPixelRGBBlend (Int32 x, Int32 y, Card32 pixel);
	
	Card32 GetPixelRGB (Int32 x, Int32 y);
	
	//	Drawing algorithms (simple lines, with or without anti-aliasing) :
	
	void DrawLine (Int32 x1, Int32 y1, Int32 x2, Int32 y2, Card32 pixel);
	void DrawAALine (Int32 x1, Int32 y1, Int32 x2, Int32 y, Card32 pixel);
	
	void DrawLine (Int32 x1, Int32 y1, Int32 x2, Int32 y2, Card32 pixel, const Card16* dash, int num);
	void DrawAALine (Int32 x1, Int32 y1, Int32 x2, Int32 y, Card32 pixel, const Card16* dash, int num);
	
	//	Fill or blend with a uniform color.
	
	void RectFill (Int32 x, Int32 y, Card32 dx, Card32 dy, Card32 pixel);
	void RectFillBlend (Int32 x, Int32 y, Card32 dx, Card32 dy, Card32 pixel);
	void RectFillBlend (Int32 x, Int32 y, Card32 dx, Card32 dy, Card32 pixel, Int32 sx, Int32 sy, GfxSurface* src_surface);
	
	void RectCopy (Int32 x, Int32 y, Card32 dx, Card32 dy, Int32 sx, Int32 sy, GfxSurface* src_surface = 0, Bool tile = TRUE);
	void RectBlend (Int32 x, Int32 y, Card32 dx, Card32 dy, Int32 sx, Int32 sy, GfxSurface* src_surface = 0, Bool tile = TRUE);
	
	void RectExtract (Int32 x, Int32 y, Card32 dx, Card32 dy, Int32 px, Int32 py, GfxSurface* dest_surface, ColorPlane plane);
	void RectExtractMono (Int32 x, Int32 y, Card32 dx, Card32 dy, Int32 px, Int32 py, Card8* mono, Card32 pitch);
	void RectInsert (Int32 x, Int32 y, Card32 dx, Card32 dy, Int32 sx, Int32 sy, GfxSurface* src_surface, ColorPlane plane);
	
	//	Utility methods used internally. They require either an RGB destination or
	//	an RGB source surface (see internal documentation).
	
	void ConvertToRGB (MemoryPtr src, MemoryPtr dst, Card32 x, Card32 dx);
	void ConvertTileToRGB (MemoryPtr src, MemoryPtr dst, Card32 x, Card32 dx);
	void ConvertFromRGB (MemoryPtr src, MemoryPtr dst, Card32 x, Card32 dx);
	void ConvertFromRGB (MemoryPtr src, MemoryPtr dst, Card32 x, Card32 dx, Card32 trans);
	void BlendWithRGB (MemoryPtr src, MemoryPtr dst, Card32 x, Card32 dx);
	void ExtractFromRGB (MemoryPtr src, MemoryPtr dst, Card32 x, Card32 dx, Card32 shift);
	void InsertIntoRGB (MemoryPtr src, MemoryPtr dst, Card32 x, Card32 dx, Card32 shift, Card32 mask);
	void ExtractMonoFromRGB (MemoryPtr src, Card8* mono, Card32 x, Card32 dx);
	void InsertMonoIntoRGB (const Card8* mono, MemoryPtr dst, Card32 x, Card32 dx);
	
	static GfxSurface* New (Card32 bpp);
	static void Delete (GfxSurface*& surface);
};

/****************************************************************/

#endif /* _GFX_GFXSURFACE_H_ */
