/*
 *	freetype.cxx
 *
 *	Provides the interface to the FreeType library.
 *
 *	(C) Copyright 1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include <OPaC/freetype.h>
#include <OPaC/new.h>

#include "../lib/freetype.h"

#include <memory.h>

/****************************************************************/

OP_IMPLEMENT_CLASS (OPFreeType);

/****************************************************************/

struct OPaC_FreeType
{
	TT_Engine			engine;
	TT_Face				face;
	TT_Instance			instance;
	TT_Glyph			glyph;
	TT_CharMap			char_map;
	TT_Glyph_Metrics	metrics;
	TT_Outline			outline;
	TT_Face_Properties	properties;
	TT_Instance_Metrics imetrics;
	TT_Raster_Map		bit;
	int					glyph_code[1024];
};

/****************************************************************/

OPFreeType::OPFreeType ()
{
	OP_METHOD (OPFreeType, 0);
	
	this->tt = (OPaC_FreeType*) malloc (sizeof (OPaC_FreeType));
	
	this->size       = 0;
	this->num_glyphs = 0;
	this->glyphs     = 0;
	
	this->is_open = FALSE;
}

OPFreeType::~OPFreeType ()
{
	OP_METHOD (~OPFreeType, 0);
	
	if (this->is_open) {
		this->CloseFace ();
	}
}

/****************************************************************/

static char virtual_palette[5] = { 0, 1, 2, 3, 4 };
static char bounded_palette[8] = { 0, 1, 2, 3, 4, 4, 4, 4 };

static TT_Error
RenderGlyph (OPaC_FreeType* tt, OPaC_TTGlyph& ttglyph,
			 int index,
			 Bool font_smoothing)
{
	TT_Load_Glyph (tt->instance, tt->glyph, index, TTLOAD_SCALE_GLYPH | TTLOAD_HINT_GLYPH);
	TT_Get_Glyph_Outline (tt->glyph, &tt->outline);
	TT_Get_Glyph_Metrics (tt->glyph, &tt->metrics);
	
	TT_F26Dot6  xmin, ymin, xmax, ymax;
	
	/* we begin by grid-fitting the bounding box */
	
	xmin = (tt->metrics.bbox.xMin+0)  & -64;
	ymin = (tt->metrics.bbox.yMin+0)  & -64;
	xmax = (tt->metrics.bbox.xMax+63) & -64;
	ymax = (tt->metrics.bbox.yMax+63) & -64;

	int dx = (xmax - xmin) / 64;
	int dy = (ymax - ymin) / 64;
	
	tt->bit.width  = dx;
	tt->bit.cols   = (dx+31) & -32;
	tt->bit.rows   = dy;
	tt->bit.flow   = TT_Flow_Down;
	tt->bit.size   = tt->bit.cols * tt->bit.rows;
	tt->bit.bitmap = OP_ALLOC_ARRAY (Card8, tt->bit.size);
	
	memset (tt->bit.bitmap, 0, tt->bit.size);
	
	(font_smoothing ? ttglyph.pixmap : ttglyph.bitmap) = (Card8*)tt->bit.bitmap;
	
	ttglyph.dx = tt->bit.width;
	ttglyph.dy = tt->bit.rows;
	ttglyph.b_width = tt->bit.cols / 8;
	ttglyph.p_width = tt->bit.cols;
	ttglyph.x_bearing = (tt->metrics.bearingX+63) / 64;
	ttglyph.x_advance = (tt->metrics.advance+63) / 64;
	ttglyph.x_min = (xmin) / 64;
	ttglyph.y_min = (ymin) / 64;
	ttglyph.x_max = (xmax) / 64;
	ttglyph.y_max = (ymax) / 64;
	
	return (font_smoothing) ? TT_Get_Glyph_Pixmap (tt->glyph, &tt->bit, -xmin, -ymin)
							: TT_Get_Glyph_Bitmap (tt->glyph, &tt->bit, -xmin, -ymin);
}


static void
CharToUnicode (OPaC_FreeType* tt)
{
	int    i, n;
	short  platform, encoding;
	
	/* First, look for a Unicode charmap */
	
	n = TT_Get_CharMap_Count (tt->face);
	
	for (i = 0; i < n; i++) {
		TT_Get_CharMap_ID (tt->face, i, &platform, &encoding);
		
		if ( (platform == 3 && encoding == 1)
		  || (platform == 0 && encoding == 0) ) {
			TT_Get_CharMap (tt->face, i, &tt->char_map );
			i = n + 1;
		}
	}
	
	if (i == n) return;
	
	for (n = 0; n < sizeof (tt->glyph_code) / sizeof (int); n++ ) {
		tt->glyph_code[n] = TT_Char_Index (tt->char_map, n);
	}
}

/****************************************************************/

Bool
OPFreeType::GetMonoBitmap (UnicodeChar index, const OPaC_TTGlyph*& pixels)
{
	if (this->is_open == FALSE) {
		return FALSE;
	}
	
	index = tt->glyph_code[index];
	
	if (index == 0) {
		return FALSE;
	}
	
	if (this->glyphs[index].bitmap == 0) {
		if (RenderGlyph (this->tt, this->glyphs[index], index, FALSE)) {
			return FALSE;
		}
	}
	
	pixels = & this->glyphs[index];
	
	return TRUE;
}

Bool
OPFreeType::GetGrayBitmap (UnicodeChar index, const OPaC_TTGlyph*& pixels)
{
	if (this->is_open == FALSE) {
		return FALSE;
	}
	
	index = tt->glyph_code[index];
	
	if (index == 0) {
		return FALSE;
	}
	
	if (this->glyphs[index].pixmap == 0) {
		if (RenderGlyph (this->tt, this->glyphs[index], index, TRUE)) {
			return FALSE;
		}
	}
	
	pixels = & this->glyphs[index];
	
	return TRUE;
}

/****************************************************************/

Bool
OPFreeType::OpenFace (const char* font_name, int size)
{
	if (this->is_open) {
		this->CloseFace ();
	}
	
	int error;
	
	error = TT_Init_FreeType (&this->tt->engine);
	if (error) return FALSE;
	
	error = TT_Open_Face (this->tt->engine, font_name, &this->tt->face);
	if (error) return FALSE;
	
	TT_Get_Face_Properties (this->tt->face, &this->tt->properties);
	
	this->num_glyphs = this->tt->properties.num_Glyphs;
	this->glyphs = OP_ALLOC_ARRAY (OPaC_TTGlyph, this->num_glyphs);
	
	memset (this->glyphs, 0, this->num_glyphs * sizeof (OPaC_TTGlyph));
	
	error = TT_New_Glyph (this->tt->face, &this->tt->glyph);
	if (error) return FALSE;
	
	error = TT_New_Instance (this->tt->face, &this->tt->instance);
	if (error) return FALSE;
	
	error = TT_Set_Instance_Resolutions (this->tt->instance, 96, 96);
	if (error) return FALSE;
	
	TT_Set_Raster_Gray_Palette (this->tt->engine, virtual_palette);
	
	this->size = size;
	
	TT_Set_Instance_CharSize (this->tt->instance, size*64);
	TT_Get_Instance_Metrics (this->tt->instance, &this->tt->imetrics);
	CharToUnicode (this->tt);
	this->tt->bit.bitmap = 0;
	this->is_open = TRUE;
	
	return TRUE;
}

Bool
OPFreeType::CloseFace ()
{	
	if (this->is_open) {
		
		this->is_open = FALSE;
		
		TT_Close_Face (this->tt->face);
		TT_Done_FreeType (this->tt->engine);
		
		OP_FREE (this->glyphs);
		
		this->glyphs     = 0;
		this->num_glyphs = 0;
		
		return TRUE;
	}
	
	return FALSE;
}

/****************************************************************/
