/*
 *	connectline.cxx
 *
 *	Draw the connecting line between two objects, possibly in
 *	two different windows.
 *
 *	(C) Copyright 1994-1998, Pierre ARNAUD, OPaC bright ideas
 *		CH-1437 SUSCEVAZ, Switzerland. All rights reserved.
 */

/******************************************************************************

  This file is part of the OPaC class library ("OPaC").

  OPaC is distributed with NO WARRANTY OF ANY KIND.  No author or distributor
  accepts any responsibility for the consequences of using it, or for whether
  it serves any particular purpose or works at all, unless he or she says so
  in writing. Refer to the OPaC Free Public License (the "License") for full
  details.

  Every copy of OPaC must include a copy of the License, normally in a plain
  ASCII text file named PUBLIC. The License grants you the right to copy,
  modify and redistribute OPaC, but only under certain conditions described
  in the License. The copyright notice and this notice must be preserved on
  all copies.

******************************************************************************/

#include "connectline.h"
#include "connectdlg.h"

#include <OPaC/dragpal.h>
#include <OPaC/event.h>
#include <OPaC/grail.h>
#include <OPaC/new.h>
#include <OPaC/pubmacro.h>
#include <OPaC/selview.h>
#include <OPaC/targetmacro.h>
#include <OPaC/view.h>
#include <OPaC/window.h>

/****************************************************************/

APP_IMPLEMENT_CLASS (OPConnectLine);

/****************************************************************/

OP_TARGET_CLASS (OPConnectLine);
OP_TARGET_ARG (OPObjectRef, object, "connection end point");
OP_TARGET_CALL (CommitConnect (object));
OP_TARGET_END ();

/****************************************************************/

#define	DOT_SIZE	8

/****************************************************************/

OPConnectLine::OPConnectLine ()
{
	this->ox = this->oy = 0;
	this->dx = this->dy = 0;
	
	this->dragging = FALSE;
	this->init_ok  = FALSE;
	
	this->target_is_window = FALSE;
	this->target_changed   = FALSE;
}


OPConnectLine::~OPConnectLine ()
{
}


void
OPConnectLine::UpdateWindows ()
{
	OPCoord x, y, d;
	
	OPCoord x1 = this->ox;
	OPCoord y1 = this->oy;
	OPCoord x2 = this->ox + this->dx;
	OPCoord y2 = this->oy + this->dy;
	
	OPCoord bx1 = this->target_box.X1 ();
	OPCoord by1 = this->target_box.Y1 ();
	OPCoord bx2 = this->target_box.X2 ();
	OPCoord by2 = this->target_box.Y2 ();
	
	x = (x2 < x1) ? (x2) : (x1);
	y = y1;
	d = (x2 < x1) ? (x1-x2) : (x2-x1);
	
	if (this->target_is_window == FALSE) {
		if ((y > by1) && (y < by2)) {
			if ((x > bx1) && (x < bx2)) {
				d -= bx2 - x;
				x  = bx2;
			} else {
				d -= x+d-bx1;
			}
		}
	}
	
	if ( (this->target_box.IsInside (x-1, y-1))
	  && (this->target_box.IsInside (x+d, y))
	  && (this->target_is_window == FALSE) ) {
		this->src_line->Hide ();
	} else {
		this->src_line->SetBox (OPRect (x-1, y-1, x+d, y));
		this->src_line->Open ();
		this->src_line->Display ();
		this->src_line->ShowNormal ();
	}
	
	x = x2;
	y = (y2 < y1) ? (y2) : (y1);
	d = (y2 < y1) ? (y1-y2) : (y2-y1);
	
	if (this->target_is_window == FALSE) {
		if ((x > bx1) && (x < bx2)) {
			if ((y > by1) && (y < by2)) {
				d -= by2 - y;
				y  = by2;
			} else {
				d -= y+d-by1;
			}
		}
	}
	
	if ( (this->target_box.IsInside (x-1, y-1))
	  && (this->target_box.IsInside (x, y+d))
	  && (this->target_is_window == FALSE) ) {
		this->dst_line->Hide ();
	} else {
		this->dst_line->SetBox (OPRect (x-1, y-1, x, y+d));
		this->dst_line->Open ();
		this->dst_line->Display ();
		this->dst_line->ShowNormal ();
	}
	
	if (this->target_changed) {
		if (this->target_is_window) {
			this->src_line->Hide ();
			this->dst_line->Hide ();
		}
	}
		
	if (this->target_view.IsValid ()) {
//		this->dst_rect->SetBox (this->target_box);
//		this->dst_rect->Open ();
//		this->dst_rect->Display ();
//		this->dst_rect->ShowNormal ();
	}
		
	if (this->target_changed) {
		if (this->target_is_window) {
			this->src_line->ShowNormal ();
			this->dst_line->ShowNormal ();
		}
	}
		
	this->target_changed = FALSE;
}


Bool
OPConnectLine::FindView (OPRectOut box, OPViewOut view)
{
	OPGrailWindowRef    grail;
	OPWindowRef         window;
	
	Count depth = 0;
	
	for (depth = 0;depth < 10; depth++) {
		
		OPCoord x = this->ox + this->dx;
		OPCoord y = this->oy + this->dy;
		
		if (OPGrailWindow::FindOwnWindow (x, y, grail, depth)) {
			
			OPCoord ox = 0;
			OPCoord oy = 0;
			
			grail->GetWindow (window);
			
			if (window.IsInvalid ())	continue;
			if (window == this->src_dot) continue;
			if (window == this->src_line) continue;
			if (window == this->dst_line) continue;
			if (window == this->dst_rect) continue;
			
			window->GetOrigin (ox, oy);
			
			x -= ox;
			y -= oy;
			
			//	Find out where the user is pointing the mouse.
			
			if (window->Detect (x, y, view)) {
				
				for (;;) {
					if (view->IsKindOf (OP_RUNTIME (OPSelView))) {
						OPSelViewRef selview = view;
						view = selview->ReturnSelectedView ();
						continue;
					}
					if (view->IsKindOf (OP_RUNTIME (OPDragPalette))) {
						OPDragPaletteRef dragpal = view;
						dragpal->GetFirstChild (view);
						continue;
					}
					break;
				}
				
				if (view.IsValid ()) {
					box = view->ReturnBox ();
					return TRUE;
				}
			}

			view = window;
			return TRUE;
		}
	}
	
	return FALSE;
}


static void
AddBlackView (OPWindowIn window, Bool just_rect = FALSE)
{
	OPConnectLineViewRef view = OP_NEW (OPConnectLineView);
	
	view->SetParent (window);
	view->SetOrigin (0, 0);
	view->SetSize (10000, 10000);
	view->DrawJustRect (just_rect);
}


void
OPConnectLine::BeginDrag (OPViewIn source, OPCoord x, OPCoord y)
{
	OPEvent::GrabWindowEvents (this->window, TRUE);
	
	this->ox = x;
	this->oy = y;
	this->dx = 0;
	this->dy = 0;

	this->src_dot  = OP_NEW (OPWindow);
	this->src_line = OP_NEW (OPWindow);
	this->dst_line = OP_NEW (OPWindow);
	this->dst_rect = OP_NEW (OPWindow);

	AddBlackView (this->src_dot);
	AddBlackView (this->src_line);
	AddBlackView (this->dst_line);
	AddBlackView (this->dst_rect, TRUE);
	
	this->src_dot->SetOrigin (x-DOT_SIZE/2, y-DOT_SIZE/2);
	this->src_dot->SetSize (DOT_SIZE, DOT_SIZE);
	this->src_dot->SetStyle (OP_WIN_NAKED | OP_WIN_TOP_LEVEL);
	this->src_dot->ShowNormal ();
	
	this->src_line->SetStyle (OP_WIN_NAKED | OP_WIN_TOP_LEVEL);
	this->dst_line->SetStyle (OP_WIN_NAKED | OP_WIN_TOP_LEVEL);
	this->dst_rect->SetStyle (OP_WIN_NAKED | OP_WIN_TOP_LEVEL | OP_WIN_TRANSPARENT);

	this->init_ok = TRUE;
	this->source_view = source;
}


void
OPConnectLine::EndDrag ()
{
	this->src_dot->Hide ();
	this->src_line->Hide ();
	this->dst_line->Hide ();
	this->dst_rect->Hide ();
	this->src_dot->Delete ();
	this->src_line->Delete ();
	this->dst_line->Delete ();
	this->dst_rect->Delete ();
	this->init_ok = FALSE;
	
	OPEvent::GrabWindowEvents (this->window, FALSE);
}


void
OPConnectLine::ButtonRelease (OPEventIn event)
{
	this->EndDrag ();
	
	if (this->target_view.IsValid ()) {
		this->target_obj = this->target_view;
		this->target_view->DragAndDrop (this->box, this, OP_DDP_CONNECT_COMMIT);
		OPConnectDialogRef  dialog_1 = OP_NEW (OPConnectDialog);
		dialog_1->Initialise (this->source_view, this->target_obj);
		this->target_view = 0;
	}
}


void
OPConnectLine::CommitConnect (OPObjectIn object)
{
	OP_METHOD (CommitConnect, 1);
	OP_ARG (1, OPObjectIn, object);
	
	this->target_obj = object;
}

void
OPConnectLine::DragNotify (OPEventIn event)
{
	if (this->init_ok == FALSE) {
		return;
	}
	
	if (this->dragging == FALSE) {
		this->dragging = TRUE;
		this->drag_ox  = this->ox - this->window->ReturnBox ().X1 ();
		this->drag_oy  = this->oy - this->window->ReturnBox ().Y1 ();
	}
	
	OPCoord x = this->ox + this->dx;
	OPCoord y = this->oy + this->dy;
	
	x += event->ReturnX () - this->drag_ox;
	y += event->ReturnY () - this->drag_oy;

	this->drag_ox = event->ReturnX ();
	this->drag_oy = event->ReturnY ();
	
	this->dx = x - this->ox;
	this->dy = y - this->oy;
	
	OPRect    box;
	OPViewRef view;
	
	if (this->FindView (box, view)) {
		if (this->target_view != view) {
			if (this->target_view.IsValid ()) {
//				this->dst_rect->Hide ();
				this->target_view->DragAndDrop (this->box, this, OP_DDP_CONNECT_LEAVE);
				this->target_view = 0;
			}
			
			OPWindowRef window;
			view->GetWindow (window);
			OPRect wbox (window->ReturnBox ());
			
			if (view == window) {
				box = wbox;
				this->target_is_window = TRUE;
			} else {
				box.SetOrigin (box.X1 () + wbox.X1 () - 2, box.Y1 () + wbox.Y1 () - 2);
				box.SetSize (box.DX () + 4, box.DY () + 4);
				this->target_is_window = FALSE;
			}
			
			this->target_view    = view;
			this->target_box     = box;
			this->target_changed = TRUE;
			
			this->target_view->DragAndDrop (this->box, this, OP_DDP_CONNECT_ENTER);
		}
	} else {
		if (this->target_view.IsValid ()) {
//			this->dst_rect->Hide ();
			this->target_view->DragAndDrop (this->box, this, OP_DDP_CONNECT_LEAVE);
			this->target_view = 0;
			this->target_box.MakeEmpty ();
		}
	}

	this->UpdateWindows ();
}
